"""change_keys_of_facters_and_violations_to_separated_table

Revision ID: 4cf3e398a81f
Revises: 21087e990aa8
Create Date: 2013-12-04 17:42:02.709200

"""

# revision identifiers, used by Alembic.
revision = '4cf3e398a81f'
down_revision = '21087e990aa8'

from alembic import op
import sqlalchemy as sa

def upgrade():
    connection = op.get_bind()

    op.create_table(
        'keys',
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('name', sa.String(2000), nullable=False)
    )

    op.add_column('facts', sa.Column('key_id', sa.Integer, nullable=False))
    op.add_column('violations', sa.Column('key_id', sa.Integer, nullable=False))

    keys = set()

    for values in connection.execute('SELECT DISTINCT(`key`) FROM violations;'):
        keys.add(values[0])

    for values in connection.execute('SELECT DISTINCT(`key`) FROM facts;'):
        keys.add(values[0])

    if keys:
        connection.execute('INSERT INTO `keys` (`name`) VALUES {0}'.format(', '.join(('(\'{0}\')'.format(key) for key in keys))))

    for values in connection.execute('SELECT id, `name` FROM `keys`'):
        connection.execute('UPDATE facts SET key_id = \'{0}\' WHERE `key` = \'{1}\''.format(*values))
        connection.execute('UPDATE violations SET key_id = \'{0}\' WHERE `key` = \'{1}\''.format(*values))

    op.create_foreign_key(
        "fk_key_fact", 'facts',
        "keys", ["key_id"], ["id"]
    )
    op.create_foreign_key(
        "fk_key_violation", 'violations',
        "keys", ["key_id"], ["id"]
    )

    op.drop_column('facts', 'key')
    op.drop_column('violations', 'key')


def downgrade():
    connection = op.get_bind()

    op.add_column('facts', sa.Column('key', sa.String(2000)))
    op.add_column('violations', sa.Column('key', sa.String(2000)))

    for values in connection.execute('SELECT id, `name` FROM `keys`'):
        id_, key = values
        connection.execute('UPDATE facts SET `key` = \'{0}\' WHERE key_id = \'{1}\''.format(key, id_))
        connection.execute('UPDATE violations SET `key` = \'{0}\' WHERE key_id = \'{1}\''.format(key, id_))

    op.drop_constraint('fk_key_fact', 'facts', type_='foreignkey')
    op.drop_column('facts', 'key_id')

    op.drop_constraint('fk_key_violation', 'violations', type_='foreignkey')
    op.drop_column('violations', 'key_id')

    op.drop_table('keys')



"""Changing facts value to Longtext

Revision ID: 3412b4b8b24e
Revises: 4779cd2391f7
Create Date: 2014-01-15 15:19:06.258099

"""

# revision identifiers, used by Alembic.
revision = '3412b4b8b24e'
down_revision = '4779cd2391f7'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.alter_column('facts', 'value', type_=sa.Text(length=4294967295), nullable=False)


def downgrade():
    op.alter_column('facts', 'value', type_=sa.Text(), nullable=False)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from derpconf.config import Config  # NOQA

MINUTE = 60
HOUR = MINUTE * 60
DAY = 24 * HOUR

Config.define('WORKER_SLEEP_TIME', 10, 'Main loop sleep time', 'Worker')
Config.define('ZOMBIE_WORKER_TIME', 200,
              'Time to remove a Worker from API List (must be greater than WORKER_SLEEP_TIME + Validation time)', 'API')

Config.define('WORKERS_LOOK_AHEAD_PAGES', 1000, 'Number of pages that will be retrieved when looking for the next job', 'Worker')

Config.define('UPDATE_PAGES_SCORE_EXPIRATION', 30, 'The expiration for lock to update pages score', 'Worker')
Config.define('UPDATE_PAGES_SCORE_SLEEP_TIME', HOUR, 'The expiration for lock to update pages score', 'Worker')

Config.define('CONNECT_TIMEOUT_IN_SECONDS', 20, 'Number of seconds a connection can take.', 'Worker')
Config.define('REQUEST_TIMEOUT_IN_SECONDS', 60, 'Number of seconds a request can take.', 'Worker')

Config.define('HOLMES_API_URL', 'http://localhost:2368', 'URL that Worker will communicate with API', 'Worker')

Config.define('LOG_LEVEL', 'ERROR', 'Default log level', 'Logging')
Config.define('LOG_FORMAT', '%(asctime)s:%(levelname)s %(module)s - %(message)s',
              'Log Format to be used when writing log messages', 'Logging')
Config.define('LOG_DATE_FORMAT', '%Y-%m-%d %H:%M:%S',
              'Date Format to be used when writing log messages.', 'Logging')

Config.define('FACTERS', [], 'List of classes to get facts about a website', 'Review')
Config.define('VALIDATORS', [], 'List of classes to validate a website', 'Review')
Config.define('REVIEW_EXPIRATION_IN_SECONDS', 6 * 60 * 60, 'Number of seconds that a review expires in.', 'Review')
Config.define('NUMBER_OF_REVIEWS_TO_KEEP', 4, 'Maximum number of reviews to keep', 'Review')

Config.define('DAYS_TO_KEEP_REQUESTS', 12, 'Number of days to keep requests', 'Requests')

Config.define('MAX_ENQUEUE_BUFFER_LENGTH', 1000,
              'Number of urls to enqueue before submitting to the /pages route', 'Validators')

# Reference data retrieved from HTTP Archive in 06-jan-2014
Config.define('MAX_IMG_REQUESTS_PER_PAGE', 40,
              'Maximum number of images per page', 'Image Request Validator')
Config.define('MAX_KB_SINGLE_IMAGE', 26,
              'Maximum size of a single image', 'Image Request Validator')
Config.define('MAX_IMG_KB_PER_PAGE', 1028,
              'Maximum size of images per page', 'Image Request Validator')
# Reference data retrieved from HTTP Archive in 06-jan-2014
Config.define('MAX_CSS_REQUESTS_PER_PAGE', 8,
              'Maximum number of external stylesheets per page', 'CSS Request Validator')
Config.define('MAX_CSS_KB_PER_PAGE_AFTER_GZIP', 46,
              'Maximum size of stylesheets per page after gzip', 'CSS Request Validator')

# Reference data retrieved from HTTP Archive in 06-jan-2014
Config.define('MAX_JS_REQUESTS_PER_PAGE', 17,
              'Maximum number of external scripts per page', 'JS Request Validator')
Config.define('MAX_JS_KB_PER_PAGE_AFTER_GZIP', 272,
              'Maximum size of scripts per page after gzip', 'JS Request Validator')

Config.define('MAX_TITLE_SIZE', 70,
              'Title tags longer than 70 characters may be truncated in the results',
              'Title Validator')
Config.define('METATAG_DESCRIPTION_MAX_SIZE', 300,
              'Description Meta tags longer than 300 characters may be truncated in the results',
              'Metatag validator')
Config.define('MAX_HEADING_HIEARARCHY_SIZE', 150,
              'Heading Hierarchy tags longer than 150 characters may be truncated in the results',
              'Heading Hierarchy Validator')
Config.define('MAX_IMAGE_ALT_SIZE', 70,
              'Image alt attributes longer than 70 characters may be truncated in the results',
              'Image Alt Attribute Validator')

Config.define('ORIGIN', '*', 'Access Control Allow Origin header value', 'Web')

Config.define('HTTP_PROXY_HOST', None, 'HTTP Proxy Host to use', 'Web')
Config.define('HTTP_PROXY_PORT', None, 'HTTP Proxy Port to use', 'Web')

Config.define('API_PROXY_HOST', None, 'HTTP Proxy Host to use to connect to the API', 'Web')
Config.define('API_PROXY_PORT', None, 'HTTP Proxy Port to use to connect to the API', 'Web')

Config.define('COMMIT_ON_REQUEST_END', True, 'Commit on request end', 'DB')

Config.define('REDISHOST', 'localhost', 'Redis host', 'Redis')
Config.define('REDISPORT', 7575, 'Redis port', 'Redis')
Config.define('REDISPASS', None, 'Redis password in case of auth', 'Redis')

Config.define('MATERIAL_GIRL_REDISPORT', 7575, 'Redis port', 'Redis')
Config.define('MATERIAL_GIRL_REDISHOST', 'localhost', 'Redis host', 'Redis')
Config.define('MATERIAL_GIRL_REDISPASS', None, 'Redis password in case of auth', 'Redis')

Config.define('REQUIRED_META_TAGS', [], 'List of required meta tags', 'Meta tag Validator')

Config.define('SCHEMA_ORG_ITEMTYPE', [], 'List of Schema.Org ItemType', 'Schema.Org ItemType')

Config.define('FORCE_CANONICAL', False, 'Force canonical', 'Force canonical')

Config.define('BLACKLIST_DOMAIN', [], 'Domain blacklist', 'Domain blacklist')

Config.define('ERROR_HANDLERS', [], 'List of classes to handle errors', 'General')

# SENTRY ERROR HANDLER
Config.define('USE_SENTRY', False, 'If set to true errors will be sent to sentry.', 'Sentry')
Config.define('SENTRY_DSN_URL', '', 'URL to use as sentry DSN.', 'Sentry')

Config.define('PAGE_COUNT_EXPIRATION_IN_SECONDS', HOUR, 'Expiration for the cache key for each domain page count', 'Cache')
Config.define('VIOLATION_COUNT_EXPIRATION_IN_SECONDS', HOUR, 'Expiration for the cache key for each domain violation count', 'Cache')
Config.define('ACTIVE_REVIEW_COUNT_EXPIRATION_IN_SECONDS', HOUR, 'Expiration for the cache key for each domain violation count', 'Cache')
Config.define('GOOD_REQUEST_COUNT_EXPIRATION_IN_SECONDS', HOUR, 'Expiration for the cache key for each domain good requests', 'Cache')
Config.define('BAD_REQUEST_COUNT_EXPIRATION_IN_SECONDS', HOUR, 'Expiration for the cache key for each domain bad requests', 'Cache')
Config.define('RESPONSE_TIME_AVG_EXPIRATION_IN_SECONDS', HOUR, 'Expiration for the cache key for each domain average response time', 'Cache')
Config.define('VIOLATIONS_BY_CATEGORY_EXPIRATION_IN_SECONDS', 6 * 60, 'Expiration for the cache key for each domain violation count by category', 'Cache')
Config.define('TOP_CATEGORY_VIOLATIONS_EXPIRATION_IN_SECONDS', 6 * 60, 'Expiration for the cache key for each domain top violation in a category', 'Cache')
Config.define('TOP_CATEGORY_VIOLATIONS_LIMIT', 10, 'Limit for the size of the list of top vilations of a key category for a domain', 'Domain Handler')
Config.define('URL_LOCK_EXPIRATION_IN_SECONDS', 30, 'Expiration for the url lock for each url', 'Cache')
Config.define('NEXT_JOB_URL_LOCK_EXPIRATION_IN_SECONDS', 3 * 60, 'Expiration for the url lock for next jobs', 'Cache')
Config.define('NEXT_JOBS_COUNT_EXPIRATION_IN_SECONDS', HOUR, 'Expiration for the cache key for next jobs count', 'Cache')
Config.define('REQUESTS_COUNT_EXPIRATION_IN_SECONDS', HOUR, 'Expiration for the cache key for requests count', 'Cache')

materials_expiration_in_seconds = {
    'domains_details': 0.5 * MINUTE + 1,
    'next_jobs_count': 13,
    'violation_count_by_category_for_domains': 3 * MINUTE + 11,
    'blacklist_domain_count': 10 * MINUTE + 1,
    'most_common_violations': HOUR + 7,
    'old_requests': (lambda config: config.get('DAYS_TO_KEEP_REQUESTS') * DAY - 1),
}
Config.define('MATERIALS_EXPIRATION_IN_SECONDS', materials_expiration_in_seconds, 'Expire times for materials', 'material')

materials_grace_period_in_seconds = {
    'domains_details': 2 * materials_expiration_in_seconds['domains_details'],
    'next_jobs_count': 2 * materials_expiration_in_seconds['next_jobs_count'],
    'violation_count_by_category_for_domains': 2 * materials_expiration_in_seconds['violation_count_by_category_for_domains'],
    'blacklist_domain_count': 2 * materials_expiration_in_seconds['blacklist_domain_count'],
    'most_common_violations': 2 * materials_expiration_in_seconds['most_common_violations'],
    'old_requests': (lambda config: 2 * materials_expiration_in_seconds['old_requests'](config)),
}
Config.define('MATERIALS_GRACE_PERIOD_IN_SECONDS', materials_grace_period_in_seconds, 'Grace period times for materials', 'material')

Config.define('DEFAULT_PAGE_SCORE', 1, 'Page Score for pages that the user includes through the UI', 'General')
Config.define('PAGE_SCORE_TAX_RATE', 0.1, 'Default tax rate for scoring pages.', 'General')
Config.define('MAX_PAGE_SCORE', 15000000, 'Maximum score of page', 'General')

Config.define('REQUEST_CACHE_EXPIRATION_IN_SECONDS', HOUR, 'Expiration in seconds for cache storage of responses.', 'Cache')

Config.define('MAX_URL_LEVELS', 20, 'Maximum levels of URL')

Config.define('GOOGLE_CLIENT_ID', None, 'Google client ID')

Config.define('LIMITER_LOCKS_EXPIRATION', 120, 'The expiration for locks in the limiter')
Config.define('LIMITER_VALUES_CACHE_EXPIRATION', 600, 'The expiration for valus in the limiter')
Config.define('DEFAULT_NUMBER_OF_CONCURRENT_CONNECTIONS', 5, 'Default number of concurrent connections', 'Limiter')

Config.define('MOST_COMMON_VIOLATIONS_CACHE_EXPIRATION', 3 * HOUR, 'Expiration for the cache key for the most common violations', 'Cache')
Config.define('MOST_COMMON_VIOLATIONS_SAMPLE_LIMIT', 50000, 'Limit for the size of the Vilation sample used in the aggregation', 'Violation Handler')

throttling_message_type = {
    'new-request': 5,
    'new-page': 2,
    'new-review': 2,
}
Config.define('EVENT_BUS_THROTTLING_MESSAGE_TYPE', throttling_message_type, 'Trottling by message type', 'Event Bus')

Config.define('SQLALCHEMY_AUTO_FLUSH', True, 'Defines whether auto-flush should be used in sqlalchemy')



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.validators.base import Validator  # NOQA

BASE_VALIDATORS = [
    'holmes.validators.title.TitleValidator',
    'holmes.validators.img_requests.ImageRequestsValidator',
    'holmes.validators.total_requests.TotalRequestsValidator',
    'holmes.validators.css_requests.CSSRequestsValidator',
    'holmes.validators.js_requests.JSRequestsValidator',
    'holmes.validators.link_crawler.LinkCrawlerValidator',
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

from sheep import Shepherd
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker
import redis
from materialgirl import Materializer
from materialgirl.storage.redis import RedisStorage

from holmes.cache import SyncCache
from holmes.utils import load_classes
from holmes.config import Config


class BaseCLI(Shepherd):
    def info(self, message):
        self.log(message, logging.info)

    def debug(self, message):
        self.log(message, logging.debug)

    def warn(self, message):
        self.log(message, logging.warn)

    def error(self, message):
        self.log(message, logging.error)

    def log(self, message, level=logging.info):
        name = self.get_description()
        level('[%s - %s] %s' % (
            name, self.parent_name, message
        ))

    def load_error_handlers(self):
        return load_classes(default=self.config.ERROR_HANDLERS)

    def get_config_class(self):
        return Config

    def connect_sqlalchemy(self):
        if getattr(self, 'db', None) is not None:
            self.db.close()

        autoflush = self.config.get('SQLALCHEMY_AUTO_FLUSH')
        connstr = self.config.SQLALCHEMY_CONNECTION_STRING
        engine = create_engine(
            connstr,
            convert_unicode=True,
            pool_size=self.config.SQLALCHEMY_POOL_SIZE,
            max_overflow=self.config.SQLALCHEMY_POOL_MAX_OVERFLOW,
            echo=self.options.verbose == 3
        )

        self.info("Connecting to \"%s\" using SQLAlchemy" % connstr)

        self.sqlalchemy_db_maker = sessionmaker(bind=engine, autoflush=autoflush, autocommit=False)
        self.db = scoped_session(self.sqlalchemy_db_maker)

    def connect_to_redis(self):
        host = self.config.get('REDISHOST')
        port = self.config.get('REDISPORT')

        self.info("Connecting to redis at %s:%d" % (host, port))
        self.redis = redis.StrictRedis(host=host, port=port, db=0)

        self.cache = SyncCache(self.db, self.redis, self.config)

        self.info("Connecting pubsub to redis at %s:%d" % (host, port))
        self.redis_pub_sub = redis.StrictRedis(host=host, port=port, db=0)

        host = self.config.get('MATERIAL_GIRL_REDISHOST')
        port = self.config.get('MATERIAL_GIRL_REDISPORT')

        self.info("Connecting material girl to redis at %s:%d" % (host, port))
        self.redis_material = redis.StrictRedis(host=host, port=port, db=0)

    def configure_material_girl(self):
        from holmes.material import configure_materials
        self.girl = Materializer(storage=RedisStorage(redis=self.redis_material))

        configure_materials(self.girl, self.db, self.config)



"""Users: email is unique

Revision ID: 4b96dd9974bb
Revises: 1d98f72f9d10
Create Date: 2014-03-14 14:40:26.069119

"""

# revision identifiers, used by Alembic.
revision = '4b96dd9974bb'
down_revision = '1d98f72f9d10'

from alembic import op


def upgrade():
    op.drop_index('idx_email', 'users')
    op.create_unique_constraint("uk_email", "users", ["email"])


def downgrade():
    op.create_index('idx_email', 'users', ['email'])
    op.drop_constraint('uk_email', 'users', type_='unique')



"""create page table

Revision ID: 2cb79bf80b46
Revises: 1fa4469fb63b
Create Date: 2013-11-13 10:23:53.306336

"""

# revision identifiers, used by Alembic.
revision = '2cb79bf80b46'
down_revision = '1fa4469fb63b'

from alembic import op
import sqlalchemy as sa


def upgrade():
    pass


def downgrade():
    pass



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.utils import get_domain_from_url, is_valid
from holmes.validators.base import Validator


class LinkWithRelNofollowValidator(Validator):
    @classmethod
    def get_links_nofollow_message(cls, value):
        return 'Links with rel="nofollow" to the same ' \
               'domain as the page make it harder for search ' \
               'engines to crawl the website. Links with ' \
               'rel="nofollow" were found for hrefs (%s).' % (
                   ', '.join([
                       '<a href="%s" target="_blank">#%s</a>' % (link, index)
                       for index, link in enumerate(value)
                    ]))

    @classmethod
    def get_violation_definitions(cls):
        return {
            'invalid.links.nofollow': {
                'title': 'Links with rel="nofollow"',
                'description': cls.get_links_nofollow_message,
                'category': 'SEO'
            }
        }

    def validate(self):

        links = self.get_links()

        page_domain, domain_url = get_domain_from_url(self.review.page_url)

        rel_nofollow = []

        for link in links:
            href = link.get('href')

            if not is_valid(href):
                continue

            link_domain, link_domain_url = get_domain_from_url(href)

            if link.get('rel') == 'nofollow' and page_domain == link_domain:
                rel_nofollow.append(href)

        if rel_nofollow:
            self.add_violation(
                key='invalid.links.nofollow',
                value=rel_nofollow,
                points=10 * len(rel_nofollow)
            )

    def get_links(self):
        return self.review.data.get('page.all_links', None)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
from uuid import uuid4
from datetime import datetime, timedelta
import hashlib
import logging

import sqlalchemy as sa
from sqlalchemy.orm import relationship
from sqlalchemy import or_
from ujson import dumps
from tornado.concurrent import return_future

from holmes.models import Base
from holmes.utils import get_domain_from_url


class Page(Base):
    __tablename__ = "pages"

    id = sa.Column(sa.Integer, primary_key=True)
    url = sa.Column('url', sa.String(2000), nullable=False)
    url_hash = sa.Column('url_hash', sa.String(128), nullable=False)
    uuid = sa.Column('uuid', sa.String(36), default=uuid4, nullable=False)
    created_date = sa.Column('created_date', sa.DateTime, default=datetime.utcnow, nullable=False)

    domain_id = sa.Column('domain_id', sa.Integer, sa.ForeignKey('domains.id'))

    reviews = relationship("Review", backref="page", foreign_keys='[Review.page_id]')

    last_review_id = sa.Column('last_review_id', sa.Integer, sa.ForeignKey('reviews.id'))
    last_review = relationship("Review", foreign_keys=[last_review_id])
    last_review_date = sa.Column('last_review_date', sa.DateTime, nullable=True)
    last_review_uuid = sa.Column('last_review_uuid', sa.String(36), nullable=True)

    last_modified = sa.Column('last_modified', sa.DateTime, nullable=True)
    expires = sa.Column('expires', sa.DateTime, nullable=True)

    violations_count = sa.Column('violations_count', sa.Integer, server_default='0', nullable=False)

    score = sa.Column('score', sa.Float, default=0.0, nullable=False)

    def to_dict(self):
        return {
            'uuid': str(self.uuid),
            'url': self.url,
            'lastModified': self.last_modified,
            'expires': self.expires,
            'score': self.score
        }

    def __str__(self):
        return str(self.uuid)

    def __repr__(self):
        return str(self)

    def get_violations_per_day(self, db):
        from holmes.models import Review, Violation  # Prevent circular dependency

        violations = db \
            .query(
                sa.func.year(Review.completed_date).label('year'),
                sa.func.month(Review.completed_date).label('month'),
                sa.func.day(Review.completed_date).label('day'),
                sa.func.count(Violation.id).label('violation_count'),
                sa.func.sum(Violation.points).label('violation_points')
            ).join(
                Page, Page.id == Review.page_id
            ).join(
                Violation, Violation.review_id == Review.id
            ).filter(Review.is_complete == True).filter(Review.page_id == self.id) \
            .group_by(
                sa.func.year(Review.completed_date),
                sa.func.month(Review.completed_date),
                sa.func.day(Review.completed_date),
            ) \
            .order_by(Review.completed_date) \
            .all()

        result = []

        for day in violations:
            dt = "%d-%d-%d" % (day.year, day.month, day.day)
            result.append({
                "completedAt": dt,
                "violation_count": int(day.violation_count),
                "violation_points": int(day.violation_points)
            })

        return result

    @classmethod
    def by_uuid(cls, uuid, db):
        return db.query(Page).filter(Page.uuid == uuid).first()

    @classmethod
    def by_url_hash(cls, url_hash, db):
        return db.query(Page).filter(Page.url_hash==url_hash).first()

    @classmethod
    def get_page_count(cls, db):
        return int(db.query(sa.func.count(Page.id)).scalar())

    @classmethod
    def update_scores(cls, individual_score, db):
        for i in range(3):
            db.begin(subtransactions=True)
            try:
                db.query(Page).update({'score': Page.score + individual_score})
                db.flush()
                db.commit()
                break
            except Exception:
                err = sys.exc_info()[1]
                if 'Deadlock found' in str(err) or 'Lock wait' in str(err):
                    logging.error('Deadlock happened! Trying again (try number %d)! (Details: %s)' % (i, str(err)))
                else:
                    db.rollback()
                    raise

    @classmethod
    def get_next_job_list(cls, db, expiration, current_page=1, page_size=200):
        from holmes.models import Domain

        lower_bound = (current_page - 1) * page_size
        upper_bound = lower_bound + page_size

        expired_time = datetime.utcnow() - timedelta(seconds=expiration)

        active_domains = Domain.get_active_domains(db)
        active_domains_ids = [item.id for item in active_domains]

        pages_query = db \
            .query(
                Page.uuid,
                Page.url,
                Page.score,
                Page.last_review_date
            ) \
            .filter(Page.domain_id.in_(active_domains_ids)) \
            .filter(or_(
                Page.last_review_date == None,
                Page.last_review_date <= expired_time
            )) \
            .order_by(Page.last_review_date.asc())

        return pages_query[lower_bound:upper_bound]

    @classmethod
    def get_next_jobs_count(cls, db, config):
        from holmes.models import Domain

        expiration = config.REVIEW_EXPIRATION_IN_SECONDS
        expired_time = datetime.utcnow() - timedelta(seconds=expiration)

        active_domains = Domain.get_active_domains(db)
        active_domains_ids = [item.id for item in active_domains]

        return db \
                .query(
                    sa.func.count(Page.id)
                ) \
                .filter(Page.domain_id.in_(active_domains_ids)) \
                .filter(or_(
                    Page.last_review_date == None,
                    Page.last_review_date <= expired_time
                )) \
                .scalar()

    @classmethod
    @return_future
    def add_page(cls, db, cache, url, score, fetch_method, publish_method, config, callback):
        domain_name, domain_url = get_domain_from_url(url)
        if not url or not domain_name:
            callback((False, url, {
                'reason': 'invalid_url',
                'url': url,
                'status': None,
                'details': 'Domain name could not be determined.'
            }))
            return

        logging.debug('Obtaining "%s"...' % url)

        fetch_method(
            url,
            cls.handle_request(cls.handle_add_page(db, cache, url, score, publish_method, config, callback)),
            proxy_host=config.HTTP_PROXY_HOST,
            proxy_port=config.HTTP_PROXY_PORT
        )

    @classmethod
    def handle_request(cls, callback):
        def handle(*args, **kw):
            response = args[-1]  # supports (url, response) and just response

            if hasattr(response, 'status_code'):
                status_code = response.status_code
            elif hasattr(response, 'code'):
                status_code = response.code
            else:
                status_code = 400

            if hasattr(response, 'body'):
                text = response.body
            elif hasattr(response, 'text'):
                text = response.text
            else:
                text = 'Empty response.text'

            callback(status_code, text, response.effective_url)

        return handle

    @classmethod
    def handle_add_page(cls, db, cache, url, score, publish_method, config, callback):
        def handle(code, body, effective_url):
            if code > 399:
                callback((False, url, {
                    'reason': 'invalid_url',
                    'url': url,
                    'status': code,
                    'details': body
                }))
                return

            if effective_url != url:
                callback((False, url, {
                    'reason': 'redirect',
                    'url': url,
                    'effectiveUrl': effective_url
                }))
                return

            domain = cls.add_domain(url, db, publish_method, config)
            page_uuid = cls.insert_or_update_page(url, score, domain, db, publish_method, cache, config)

            callback((True, url, page_uuid))

        return handle

    @classmethod
    def insert_or_update_page(cls, url, score, domain, db, publish_method, cache, config):
        url = url.encode('utf-8')
        url_hash = hashlib.sha512(url).hexdigest()
        page = Page.by_url_hash(url_hash, db)

        if page:
            cache.increment_page_score(page.url)
            return page.uuid

        try:
            page_uuid = uuid4()
            query_params = {
                'url': url,
                'url_hash': url_hash,
                'uuid': page_uuid,
                'domain_id': domain.id,
                'created_date': datetime.utcnow(),
                'score': score
            }

            db.execute(
                'INSERT INTO pages (url, url_hash, uuid, domain_id, created_date, score) ' \
                'VALUES (:url, :url_hash, :uuid, :domain_id, :created_date, :score) ON DUPLICATE KEY ' \
                'UPDATE score = :score',
                query_params
            )

            cache.increment_page_count(domain)
            cache.increment_page_count()
            cache.increment_next_jobs_count()

        except Exception:
            err = sys.exc_info()[1]
            if 'Duplicate entry' in str(err):
                logging.debug('Duplicate entry! (Details: %s)' % str(err))
            else:
                raise

        publish_method(dumps({
            'type': 'new-page',
            'pageUrl': str(url)
        }))

        return page_uuid

    @classmethod
    def add_domain(cls, url, db, publish_method, config):
        from holmes.models import Domain

        domain_name, domain_url = get_domain_from_url(url)

        domains = db.query(Domain).filter(or_(
            Domain.name == domain_name,
            Domain.name == domain_name.rstrip('/'),
            Domain.name == "%s/" % domain_name
        )).all()

        if not domains:
            domain = None
        else:
            domain = domains[0]

        if not domain:
            url_hash = hashlib.sha512(domain_url).hexdigest()
            domain = Domain(url=domain_url, url_hash=url_hash, name=domain_name)
            db.add(domain)
            db.flush()

            publish_method(dumps({
                'type': 'new-domain',
                'domainUrl': str(domain_url)
            }))

            from holmes.models import Limiter
            connections = config.DEFAULT_NUMBER_OF_CONCURRENT_CONNECTIONS
            Limiter.add_or_update_limiter(db, domain_url, connections)

        return domain

    @classmethod
    def update_pages_score(cls, db, cache, config):
        pages = cache.seized_pages_score()

        for page_id, score in pages:
            db \
                .query(Page).filter(Page.id == page_id) \
                .update(
                    {'score': sa.func.least(
                        config.MAX_PAGE_SCORE, Page.score + score)
                    },
                    synchronize_session=False
                )



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.validators.base import Validator


class MetaRobotsValidator(Validator):
    META_ROBOTS_NO_INDEX = 'A meta tag with the robots="noindex" ' \
                           'attribute tells the search engines that ' \
                           'they should not index this page.'

    META_ROBOTS_NO_FOLLOW = 'A meta tag with the robots="nofollow" ' \
                            'attribute tells the search engines that they ' \
                            'should not follow any links in this page.'

    @classmethod
    def get_violation_definitions(cls):
        return {
            'presence.meta.robots.noindex': {
                'title': 'Meta Robots with value of noindex',
                'description': lambda value: cls.META_ROBOTS_NO_INDEX,
                'category': 'SEO'
            },
            'presence.meta.robots.nofollow': {
                'title': 'Meta Robots with value of nofollow',
                'description': lambda value: cls.META_ROBOTS_NO_FOLLOW,
                'category': 'SEO'
            }
        }

    def validate(self):
        tags = self.get_meta_tags()

        for tag in tags:
            if tag['key'] == 'robots' and tag['content'] == 'noindex':
                self.add_violation(
                    key='presence.meta.robots.noindex',
                    value=None,
                    points=80
                )

            if tag['key'] == 'robots' and tag['content'] == 'nofollow':
                self.add_violation(
                    key='presence.meta.robots.nofollow',
                    value=None,
                    points=50
                )

    def get_meta_tags(self):
        return self.review.data.get('meta.tags', None)



"""create domains table

Revision ID: 34a477a7f285
Revises: None
Create Date: 2013-11-13 09:49:16.883664

"""

# revision identifiers, used by Alembic.
revision = '34a477a7f285'
down_revision = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'domains',
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('url', sa.String(2000), nullable=False),
        sa.Column('name', sa.String(2000), nullable=False),
    )


def downgrade():
    op.drop_table('domains')



#!/usr/bin/python
# -*- coding: utf-8 -*-

import pkgutil
import pkg_resources

from raven import Client


def get_modules():
    resolved = {}
    modules = [mod[1] for mod in tuple(pkgutil.iter_modules())]
    for module in modules:
        try:
            res_mod = pkg_resources.get_distribution(module)
            if res_mod is not None:
                resolved[module] = res_mod.version
        except pkg_resources.DistributionNotFound:
            pass

    return resolved


class SentryErrorHandler(object):
    def __init__(self, config):
        self.config = config

        if self.config.USE_SENTRY:
            self.sentry = Client(self.config.SENTRY_DSN_URL)

        self.modules = get_modules()

    def handle_exception(self, typ, value, tb, extra={}):
        if self.config.USE_SENTRY:
            self.sentry.captureException(
                (typ, value, tb),
                extra=extra,
                data={
                    'modules': self.modules
                }
            )



"""add last_review column to pages table

Revision ID: 1a316b08d134
Revises: 3283f8a20fc9
Create Date: 2013-11-13 15:37:58.838122

"""

# revision identifiers, used by Alembic.
revision = '1a316b08d134'
down_revision = '3283f8a20fc9'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        'pages',
        sa.Column('last_review_id', sa.Integer)
    )

    op.create_foreign_key(
        "fk_page_last_review", "pages",
        "reviews", ["last_review_id"], ["id"]
    )


def downgrade():
    op.drop_constraint('fk_page_last_review', 'pages', type_='foreignkey')
    op.drop_column('pages', 'last_review_id')



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.facters import Facter


class MetaTagsFacter(Facter):

    @classmethod
    def get_fact_definitions(cls):
        return {
            'meta.tags': {
                'title': 'Meta Tags',
                'description': lambda value: value,
                'unit': 'values',
                'category': 'Meta',
            }
        }

    def get_facts(self):

        values = self.process_meta_tags()

        self.add_fact(
            key='meta.tags',
            value=values,
        )

        self.review.data['meta.tags'] = values

    def process_meta_tags(self):
        """
        Case 1: {'charset': 'utf-8'}

        KEY => 'charset'
        CONTENT => 'utf-8'
        PROPERTY =>


        Case 2: {"content":"website","property":"og:type"}

        KEY => 'og:type'
        CONTENT => 'website'
        PROPERTY => 'property'


        Case 3: {'content': 'text\\/html;charset=UTF-8',
                 'http-equiv': 'Content-Type'}

        KEY => 'Content-Type'
        CONTENT => 'text\\/html;charset=UTF-8'
        PROPERTY => 'http-equiv'
        """

        data = []
        for tag in self.get_meta_tags():
            if len(tag) == 1:
                key = tag.iterkeys().next()
                content = tag.itervalues().next()
                prop = None
            elif len(tag) > 1:
                if 'content' in tag:
                    content = tag.get('content')
                    tag.pop('content')
                else:
                    content = None
                key = tag.itervalues().next()
                prop = tag.iterkeys().next()
            data.append({'key': key, 'content': content, 'property': prop})
        return data

    def get_meta_tags(self):
        meta_tags = self.reviewer.current_html.cssselect('meta')
        values = []
        for tags in meta_tags:
            values.append(dict(tags.items()))
        return values



#!/usr/bin/python
# -*- coding: utf-8 -*-

from gzip import GzipFile
from cStringIO import StringIO
from datetime import datetime, timedelta
import logging

import msgpack
from tornado.concurrent import return_future
from ujson import loads, dumps
from octopus.model import Response
from sqlalchemy import or_
from retools.lock import Lock, LockTimeout

from holmes.models import Domain, Page, Limiter, Violation, Request


class Cache(object):
    def __init__(self, application):
        self.application = application
        self.redis = self.application.redis
        self.db = self.application.db
        self.config = self.application.config

    def get_domain_name(self, domain_name):
        if isinstance(domain_name, Domain):
            return domain_name.name

        return domain_name or 'page'

    @return_future
    def has_key(self, key, callback):
        self.redis.exists(key, callback)

    @return_future
    def increment_violations_count(self, domain_name, increment=1, callback=None):
        self.increment_count(
            'violation-count',
            domain_name,
            lambda domain: domain.get_violation_data(self.db),
            increment,
            callback
        )

    @return_future
    def increment_active_review_count(self, domain_name, increment=1, callback=None):
        self.increment_count(
            'active-review-count',
            domain_name,
            lambda domain: domain.get_active_review_count(self.db),
            increment,
            callback
        )

    @return_future
    def increment_page_count(self, domain_name=None, increment=1, callback=None):
        self.increment_count(
            'page-count',
            domain_name,
            lambda domain: domain.get_page_count(self.db),
            increment,
            callback
        )

    def increment_count(self, key, domain_name, get_default_method, increment=1, callback=None):
        key = '%s-%s' % (self.get_domain_name(domain_name), key)
        self.has_key(key, self.handle_has_key(key, domain_name, get_default_method, increment, callback))

    def handle_has_key(self, key, domain_name, get_default_method, increment, callback):
        def handle(has_key):
            domain = domain_name
            if domain and not isinstance(domain, Domain):
                domain = Domain.get_domain_by_name(domain_name, self.db)

            #if not domain:
                #callback(None)
                #return

            if has_key:
                self.redis.incrby(key, increment, callback=callback)
            else:
                if domain is None:
                    value = Page.get_page_count(self.db) + increment - 1
                else:
                    value = get_default_method(domain) + increment - 1

                self.redis.set(key, value, callback=callback)

        return handle

    @return_future
    def increment_next_jobs_count(self, increment=1, callback=None):
        self.increment_data(
            'next-jobs',
            lambda: Page.get_next_jobs_count(self.db, self.config),
            increment,
            callback
        )

    @return_future
    def increment_requests_count(self, increment=1, callback=None):
        self.increment_data(
            'requests-count',
            lambda: Request.get_requests_count(self.db),
            increment,
            callback
        )

    def increment_data(self, key, get_default_method, increment=1, callback=None):
        self.has_key(key, self.data_handle_has_key(key, get_default_method, increment, callback))

    def data_handle_has_key(self, key, get_default_method, increment, callback):
        def handle(has_key):
            if has_key:
                self.redis.incrby(key, increment, callback=callback)
            else:
                value = get_default_method() + increment
                self.redis.set(key, value, callback=callback)

        return handle

    @return_future
    def get_page_count(self, domain_name=None, callback=None):
        self.get_count(
            'page-count',
            domain_name,
            int(self.config.PAGE_COUNT_EXPIRATION_IN_SECONDS),
            lambda domain: domain.get_page_count(self.db),
            callback=callback
        )

    @return_future
    def get_violation_count(self, domain_name, callback=None):
        self.get_count(
            'violation-count',
            domain_name,
            int(self.config.PAGE_COUNT_EXPIRATION_IN_SECONDS),
            lambda domain: domain.get_violation_data(self.db),
            callback=callback
        )

    @return_future
    def get_active_review_count(self, domain_name, callback=None):
        self.get_count(
            'active-review-count',
            domain_name,
            int(self.config.ACTIVE_REVIEW_COUNT_EXPIRATION_IN_SECONDS),
            lambda domain: domain.get_active_review_count(self.db),
            callback=callback
        )

    @return_future
    def get_good_request_count(self, domain_name, callback=None):
        self.get_count(
            'good-request-count',
            domain_name,
            int(self.config.GOOD_REQUEST_COUNT_EXPIRATION_IN_SECONDS),
            lambda domain: domain.get_good_request_count(self.db),
            callback=callback
        )

    @return_future
    def get_bad_request_count(self, domain_name, callback=None):
        self.get_count(
            'bad-request-count',
            domain_name,
            int(self.config.BAD_REQUEST_COUNT_EXPIRATION_IN_SECONDS),
            lambda domain: domain.get_bad_request_count(self.db),
            callback=callback
        )

    @return_future
    def get_response_time_avg(self, domain_name, callback=None):
        self.get_avg(
            'response-time-avg',
            domain_name,
            int(self.config.RESPONSE_TIME_AVG_EXPIRATION_IN_SECONDS),
            lambda domain: domain.get_response_time_avg(self.db),
            callback=callback
        )

    @return_future
    def get_top_in_category_for_domain(self, domain, key_category_id, limit, callback=None):
        self.get_data(
            '%s-top-violations-cat-%s' % (domain.name, key_category_id),
            int(self.config.TOP_CATEGORY_VIOLATIONS_EXPIRATION_IN_SECONDS),
            lambda: Violation.get_top_in_category_for_domain(self.db, domain, key_category_id, limit),
            callback=callback
        )

    @return_future
    def get_most_common_violations(self, violation_definitions, sample_limit, callback=None):
        self.get_data(
            'most-common-violations',
            int(self.config.MOST_COMMON_VIOLATIONS_CACHE_EXPIRATION),
            lambda: Violation.get_most_common_violations(self.db, violation_definitions, sample_limit),
            callback=callback
        )

    @return_future
    def get_next_jobs_count(self, callback=None):
        self.get_data(
            'next-jobs',
            int(self.config.NEXT_JOBS_COUNT_EXPIRATION_IN_SECONDS),
            lambda: Page.get_next_jobs_count(self.db, self.config),
            callback=callback
        )

    @return_future
    def get_requests_count(self, callback=None):
        self.get_data(
            'requests-count',
            int(self.config.REQUESTS_COUNT_EXPIRATION_IN_SECONDS),
            lambda: Request.get_requests_count(self.db),
            callback=callback
        )

    def get_domain(self, domain_name):
        domain = domain_name
        if domain and not isinstance(domain, Domain):
            domain = Domain.get_domain_by_name(domain_name, self.db)
        return domain

    def get_count(self, key, domain_name, expiration, get_count_method, callback=None):
        cache_key = '%s-%s' % (self.get_domain_name(domain_name), key)
        self.redis.get(cache_key, callback=self.handle_get_count(key, domain_name, expiration, get_count_method, callback))

    def handle_get_count(self, key, domain_name, expiration, get_count_method, callback):
        def handle(count):
            if count is not None:
                callback(int(count))
                return

            domain = self.get_domain(domain_name)

            if domain is None:
                count = Page.get_page_count(self.db)
            else:
                count = get_count_method(domain)

            cache_key = '%s-%s' % (self.get_domain_name(domain), key)

            self.redis.setex(
                key=cache_key,
                value=int(count),
                seconds=expiration,
                callback=self.handle_set_count(count, callback)
            )

        return handle

    def handle_set_count(self, count, callback):
        def handle(*args, **kw):
            callback(count)

        return handle

    def get_avg(self, key, domain_name, expiration, get_avg_method, callback=None):
        cache_key = '%s-%s' % (self.get_domain_name(domain_name), key)
        self.redis.get(cache_key, callback=self.handle_get_avg(key, domain_name, expiration, get_avg_method, callback))

    def handle_get_avg(self, key, domain_name, expiration, get_avg_method, callback):
        def handle(avg):
            if avg is not None:
                callback(float(avg))
                return

            domain = self.get_domain(domain_name)

            avg = get_avg_method(domain)

            cache_key = '%s-%s' % (self.get_domain_name(domain), key)

            self.redis.setex(
                key=cache_key,
                value=float(avg),
                seconds=expiration,
                callback=self.handle_set_avg(avg, callback)
            )

        return handle

    def handle_set_avg(self, avg, callback):
        def handle(*args, **kw):
            callback(avg)

        return handle

    def get_data(self, key, expiration, get_data_method, callback=None):
        self.redis.get(key, callback=self.handle_get_data(key, expiration, get_data_method, callback))

    def handle_get_data(self, key, expiration, get_data_method, callback):
        def handle(data):
            if data is not None:
                callback(loads(data))
                return

            data = get_data_method()

            self.redis.setex(
                key=key,
                value=dumps(data),
                seconds=expiration,
                callback=self.handle_set_data(data, callback)
            )

        return handle

    def handle_set_data(self, data, callback):
        def handle(*args, **kw):
            callback(data)

        return handle

    @return_future
    def lock_page(self, url, callback=None):
        expiration = self.config.URL_LOCK_EXPIRATION_IN_SECONDS

        self.redis.setex(
            key='%s-lock' % url,
            value=1,
            seconds=expiration,
            callback=callback
        )

    @return_future
    def has_lock(self, url, callback=None):
        self.redis.get(
            key='%s-lock' % url,
            callback=self.handle_get_lock_page(url, callback)
        )

    def handle_get_lock_page(self, url, callback):
        def handle(value):
            callback(value == '1')

        return handle

    @return_future
    def release_lock_page(self, url, callback):
        self.redis.delete('%s-lock' % url, callback=callback)

    @return_future
    def get_limit_usage(self, url, callback):
        self.redis.zcard('limit-for-%s' % url, callback=callback)

    @return_future
    def get_limit_usage_by_domain(self, domain_url, callback):
        self.redis.keys('limit-for-%s*' % domain_url, callback)

    @return_future
    def delete_limit_usage_by_domain(self, domain_url, callback):
        self.get_limit_usage_by_domain(
            domain_url,
            callback=self.handle_delete_limit_usage_by_domain(domain_url, callback)
        )

    def handle_delete_limit_usage_by_domain(self, domain_url, callback):
        def handle(keys):
            self.redis.delete(keys)
            callback()
        return handle

    @return_future
    def remove_domain_limiters_key(self, callback):
        self.redis.delete('domain-limiters', callback=callback)

    @return_future
    def increment_page_score(self, page_id, increment=1, callback=None):
        self.redis.zincrby('page-scores', increment, page_id, callback=callback)


class SyncCache(object):
    def __init__(self, db, redis, config):
        self.db = db
        self.redis = redis
        self.config = config

    def has_key(self, key):
        return self.redis.exists(key)

    def get_domain_name(self, domain_name):
        if isinstance(domain_name, Domain):
            return domain_name.name

        return domain_name or 'page'

    def increment_violations_count(self, domain_name, increment=1):
        self.increment_count(
            'violation-count',
            domain_name,
            lambda domain: domain.get_violation_data(self.db),
            increment
        )

    def increment_active_review_count(self, domain_name, increment=1):
        self.increment_count(
            'active-review-count',
            domain_name,
            lambda domain: domain.get_active_review_count(self.db),
            increment,
        )

    def increment_page_count(self, domain_name=None, increment=1):
        self.increment_count(
            'page-count',
            domain_name,
            lambda domain: domain.get_page_count(self.db),
            increment,
        )

    def increment_count(self, key, domain_name, get_default_method, increment=1):
        key = '%s-%s' % (self.get_domain_name(domain_name), key)

        has_key = self.has_key(key)

        domain = domain_name
        if domain and not isinstance(domain, Domain):
            domain = Domain.get_domain_by_name(domain_name, self.db)

        if has_key:
            self.redis.incrby(key, increment)
        else:
            if domain is None:
                value = Page.get_page_count(self.db) + increment - 1
            else:
                value = get_default_method(domain) + increment - 1

            self.redis.set(key, value)

    def increment_next_jobs_count(self, increment=1):
        self.increment_data(
            'next-jobs',
            lambda: Page.get_next_jobs_count(self.db, self.config),
            increment
        )

    def increment_requests_count(self, increment=1):
        self.increment_data(
            'requests-count',
            lambda: Request.get_requests_count(self.db),
            increment
        )

    def increment_data(self, key, get_default_method, increment=1):
        has_key = self.has_key(key)

        if has_key:
            self.redis.incrby(key, increment)
        else:
            value = get_default_method() + increment
            self.redis.set(key, value)

    def get_page_count(self, domain_name=None):
        return self.get_count(
            'page-count',
            domain_name,
            int(self.config.PAGE_COUNT_EXPIRATION_IN_SECONDS),
            lambda domain: domain.get_page_count(self.db),
        )

    def get_violation_count(self, domain_name):
        return self.get_count(
            'violation-count',
            domain_name,
            int(self.config.PAGE_COUNT_EXPIRATION_IN_SECONDS),
            lambda domain: domain.get_violation_data(self.db),
        )

    def get_active_review_count(self, domain_name):
        return self.get_count(
            'active-review-count',
            domain_name,
            int(self.config.ACTIVE_REVIEW_COUNT_EXPIRATION_IN_SECONDS),
            lambda domain: domain.get_active_review_count(self.db)
        )

    def get_count(self, key, domain_name, expiration, get_count_method):
        cache_key = '%s-%s' % (self.get_domain_name(domain_name), key)

        count = self.redis.get(cache_key)

        if count is not None:
            return int(count)

        domain = domain_name
        if domain and not isinstance(domain, Domain):
            domain = Domain.get_domain_by_name(domain_name, self.db)

        if domain is None:
            count = Page.get_page_count(self.db)
        else:
            count = get_count_method(domain)

        cache_key = '%s-%s' % (self.get_domain_name(domain), key)

        self.redis.setex(
            cache_key,
            expiration,
            value=int(count)
        )

        return int(count)

    def get_request(self, url):
        cache_key = "urls-%s" % url

        contents = self.redis.get(cache_key)

        if not contents:
            return url, None

        item = msgpack.unpackb(contents)

        text = GzipFile(mode='r', fileobj=StringIO(item['body'])).read()

        response = Response(
            url=url,
            status_code=item['status_code'],
            headers=item['headers'],
            cookies=item['cookies'],
            text=text,
            effective_url=item['effective_url'],
            error=item['error'],
            request_time=float(item['request_time'])
        )

        response.from_cache = True

        return url, response

    def set_request(self, url, status_code, headers, cookies, text, effective_url, error, request_time, expiration):
        if status_code > 399 or status_code < 100:
            return

        cache_key = "urls-%s" % url

        out = StringIO()
        with GzipFile(fileobj=out, mode="w") as f:
            f.write(text)
        text = out.getvalue()

        value = msgpack.packb({
            'url': url,
            'body': text,
            'status_code': status_code,
            'headers': headers,
            'cookies': cookies,
            'effective_url': effective_url,
            'error': error,
            'request_time': request_time
        })

        self.redis.setex(
            cache_key,
            expiration,
            value,
        )

    def lock_next_job(self, url, expiration):
        return self.redis.lock('%s-next-job-lock' % url, expiration)

    def has_next_job_lock(self, url, expiration):
        lock = self.lock_next_job(url, expiration)
        has_acquired = lock.acquire(blocking=False)
        if not has_acquired:
            return None
        return lock

    def release_next_job(self, lock):
        return lock.release()

    def set_domain_limiters(self, domains, expiration):
        self.redis.setex(
            'domain-limiters',
            expiration,
            dumps(domains)
        )

    def get_domain_limiters(self):
        domains = self.redis.get('domain-limiters')

        if domains:
            domains = loads(domains)
        else:
            limiters = Limiter.get_all(self.db)
            if limiters:
                domains = [{d.url: d.value} for d in limiters]
                self.set_domain_limiters(
                    domains,
                    self.config.LIMITER_VALUES_CACHE_EXPIRATION
                )

        return domains

    def get_limit_usage(self, url):
        return self.redis.zcard('limit-for-%s' % url)

    def increment_page_score(self, page_id, increment=1):
        self.redis.zincrby('page-scores', page_id, increment)

    def seized_pages_score(self):
        pages = self.redis.zrange('page-scores', 0, -1, withscores=True)
        self.redis.zremrangebyrank('page-scores', 0, -1)
        return pages

    def lock_update_pages_score(self, expiration):
        return self.redis.lock('update-pages-score-lock', expiration)

    def has_update_pages_lock(self, expiration):
        lock = self.lock_update_pages_score(expiration)
        has_acquired = lock.acquire(blocking=False)
        if not has_acquired:
            return None
        return lock

    def release_update_pages_lock(self, lock):
        return lock.release()

    def get_limiter_buckets(self, active_domains, avg_links_per_page=10.0):
        available = []
        all_limiters = reversed(sorted(Limiter.get_limiters_for_domains(self.db, active_domains), key=lambda item: item.url))

        for limiter in all_limiters:
            capacity = float(limiter.value - self.get_limit_usage(limiter.url))
            available.append((limiter, capacity))

        return available

    def fill_job_bucket(self, expiration, look_ahead_pages=1000, avg_links_per_page=10.0):
        try:
            with Lock('next-job-fill-bucket-lock', redis=self.redis):
                logging.info('Refilling job bucket. Lock acquired...')
                expired_time = datetime.utcnow() - timedelta(seconds=expiration)

                active_domains = Domain.get_active_domains(self.db)

                if not active_domains:
                    return

                active_domains_ids = [item.id for item in active_domains]

                limiter_buckets = self.get_limiter_buckets(active_domains, avg_links_per_page)

                all_domains_pages_in_need_of_review = []

                for domain_id in active_domains_ids:
                    pages = self.db \
                        .query(
                            Page.uuid,
                            Page.url,
                            Page.score,
                            Page.last_review_date
                        ) \
                        .filter(Page.domain_id == domain_id) \
                        .filter(or_(
                            Page.last_review_date == None,
                            Page.last_review_date <= expired_time
                        )) \
                        .order_by(Page.last_review_date.asc())[:look_ahead_pages]

                    if pages:
                        all_domains_pages_in_need_of_review.append(pages)

                logging.debug('Total of %d pages found to add to redis.' % (sum([len(item) for item in all_domains_pages_in_need_of_review])))

                item_count = int(self.redis.scard('next-job-bucket'))
                current_domain = 0
                while item_count < look_ahead_pages and len(all_domains_pages_in_need_of_review) > 0:
                    if current_domain >= len(all_domains_pages_in_need_of_review):
                        current_domain = 0

                    item = all_domains_pages_in_need_of_review[current_domain].pop(0)

                    has_limit = True
                    logging.debug('Available Limit Buckets: %s' % limiter_buckets)
                    for index, (limit, available) in enumerate(limiter_buckets):
                        if limit.matches(item.url):
                            if available <= 0:
                                has_limit = False
                                break
                            limiter_buckets[index] = (limit, available - 1)

                    if has_limit:
                        self.redis.sadd('next-job-bucket', dumps({
                            'page': str(item.uuid),
                            'url': item.url
                        }))

                        item_count += 1

                    # if there are not any more pages in this domain remove it from dictionary
                    if not all_domains_pages_in_need_of_review[current_domain]:
                        del all_domains_pages_in_need_of_review[current_domain]

                    current_domain += 1

                logging.debug('ADDED A TOTAL of %d ITEMS TO REDIS...' % item_count)

        except LockTimeout:
            logging.info("Can't acquire lock. Moving on...")

    def get_next_job(self, expiration, look_ahead_pages=1000):
        logging.info('Getting next job from the bucket...')
        item = self.redis.spop('next-job-bucket')

        job_bucket_count = self.redis.scard('next-job-bucket')
        if job_bucket_count < look_ahead_pages * 0.1:
            logging.info('Bucket near empty (%d items). Must refill...' % job_bucket_count)
            self.fill_job_bucket(expiration, look_ahead_pages)

        if item is None:
            return None

        item = loads(item)
        logging.debug('Next job found: %s' % item['url'])

        return item



#!/usr/bin/python
# -*- coding: utf-8 -*-

from tornado import gen
from functools import partial

from holmes.handlers import BaseHandler
from holmes.models import Review, Violation


class MostCommonViolationsHandler(BaseHandler):

    @gen.coroutine
    def get(self):
        violations = dict(self.girl.get('most_common_violations'))

        result = []
        for violation in self.application.violation_definitions.values():
            result.append({
                'name': violation['title'],
                'key': violation['key'].name,
                'category': violation['category'],
                'count': violations.get(violation['key'].name, 0),
            })

        self.write_json(result)
        self.finish()


class ViolationHandler(BaseHandler):

    @gen.coroutine
    def get(self, key_name):
        current_page = int(self.get_argument('current_page', 1))
        page_size = int(self.get_argument('page_size', 10))
        domain_filter = self.get_argument('domain_filter', None)
        page_filter = self.get_argument('page_filter', None)

        violations = self.application.violation_definitions
        violation_title = violations[key_name]['title']
        key_id = violations[key_name]['key'].id

        reviews = Review.get_by_violation_key_name(
            self.db,
            key_id,
            current_page=current_page,
            page_size=page_size,
            domain_filter=domain_filter,
            page_filter=page_filter,
        )

        reviews_count = Review.count_by_violation_key_name(
            self.db,
            key_id,
            domain_filter=domain_filter,
            page_filter=page_filter
        )

        reviews_data = []
        for item in reviews:
            reviews_data.append({
                'uuid': item.review_uuid,
                'page': {
                    'uuid': item.page_uuid,
                    'url': item.url,
                    'completedAt': item.completed_date
                }
            })

        violation = {
            'title': violation_title,
            'reviews': reviews_data,
            'reviewsCount': reviews_count
        }

        self.write_json(violation)
        self.finish()


class ViolationDomainsHandler(BaseHandler):

    def __init__(self, *args, **kw):
        super(ViolationDomainsHandler, self).__init__(*args, **kw)
        self.key_details_handler = {
            'blacklist.domains': partial(self.girl.get, 'blacklist_domain_count')
        }

    @gen.coroutine
    def get(self, key_name):
        violations = self.application.violation_definitions
        violation_title = violations[key_name]['title']
        key_id = violations[key_name]['key'].id

        domains = Violation.get_by_key_id_group_by_domain(self.db, key_id)

        violation = {
            'title': violation_title,
            'domains': [{'name': name, 'count': count} for (name, count) in domains],
            'total': sum(count for (name, count) in domains)
        }

        if key_name in self.key_details_handler:
            violation['details'] = self.key_details_handler[key_name]()

        self.write_json(violation)
        self.finish()


class ViolationsHandler(BaseHandler):
    def get(self):
        violations = self.application.violation_definitions

        json = []
        for key in violations.keys():
            violation = violations.get(key)
            json.append({
                'key_name': key,
                'title': violation.get('title'),
                'category': violation.get('category', None)
            })

        self.write_json(json)
        self.finish()



#!/usr/bin/python
# -*- coding: utf-8 -*-

import sqlalchemy as sa
from sqlalchemy import func
from datetime import date, datetime, timedelta

from holmes.utils import get_status_code_title
from holmes.models import Base


class Request(Base):
    __tablename__ = "requests"

    id = sa.Column(sa.Integer, primary_key=True)
    domain_name = sa.Column('domain_name', sa.String(120), nullable=False)
    url = sa.Column('url', sa.Text(), nullable=False)
    effective_url = sa.Column('effective_url', sa.Text(), nullable=False)
    status_code = sa.Column('status_code', sa.Integer, nullable=False)
    response_time = sa.Column('response_time', sa.Float, nullable=False)
    completed_date = sa.Column('completed_date', sa.Date, nullable=False)
    review_url = sa.Column('review_url', sa.Text(), nullable=False)

    def to_dict(self):
        return {
            'domain_name': str(self.domain_name),
            'url': self.url,
            'effective_url': self.effective_url,
            'status_code': self.status_code,
            'response_time': self.response_time,
            'completed_date': self.completed_date,
            'review_url': self.review_url
        }

    def __str__(self):
        return "%s (%s)" % (self.url, self.status_code)

    def __repr__(self):
        return str(self)

    @classmethod
    def get_status_code_info(self, domain_name, db):
        result = []

        query = db \
            .query(
                Request.status_code,
                sa.func.count(Request.status_code).label('total')
            ) \
            .filter(Request.domain_name == domain_name) \
            .group_by(Request.status_code) \
            .all()

        for i in query:
            result.append({
                'code': i.status_code,
                'title': get_status_code_title(i.status_code),
                'total': i.total
            })

        return result

    @classmethod
    def get_requests_by_status_code(self, domain_name, status_code, db, current_page=1, page_size=10):
        lower_bound = (current_page - 1) * page_size
        upper_bound = lower_bound + page_size

        requests = db \
            .query(Request.id, Request.url, Request.review_url, Request.completed_date) \
            .filter(Request.domain_name == domain_name) \
            .filter(Request.status_code == status_code) \
            .order_by('completed_date desc')[lower_bound:upper_bound]

        return requests

    @classmethod
    def get_requests_by_status_count(self, domain_name, status_code, db):
        return db \
            .query(func.count(Request.id)) \
            .filter(Request.domain_name == domain_name) \
            .filter(Request.status_code == status_code) \
            .scalar()

    @classmethod
    def get_requests_count(cls, db):
        return int(db.query(sa.func.count(Request.id)).scalar())

    @classmethod
    def get_last_requests(self, db, current_page=1, page_size=10):
        lower_bound = (current_page - 1) * page_size
        upper_bound = lower_bound + page_size

        return db \
            .query(Request) \
            .order_by('id desc')[lower_bound:upper_bound]

    @classmethod
    def get_requests_count_by_status_in_period_of_days(self, db, from_date, to_date=None):
        if to_date is None:
            to_date = datetime.utcnow()

        return db \
            .query(Request.status_code, sa.func.count(Request.id).label('count')) \
            .filter(Request.completed_date.between(from_date.date(), to_date.date())) \
            .group_by(Request.status_code) \
            .order_by('count DESC') \
            .all()

    @classmethod
    def delete_old_requests(self, db, config):
        dt = date.today() - timedelta(days=config.DAYS_TO_KEEP_REQUESTS)

        return db \
            .query(Request) \
            .filter(Request.completed_date <= dt) \
            .delete()



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.validators.base import Validator


class HeadingHierarchyValidator(Validator):

    @classmethod
    def get_violation_description(cls, value):
        return (
            'Heading hierarchy values bigger than %s characters aren`t good '
            'for Search Engines. This elements were found:'
            '<ul class="violation-hh-list">%s</ul>' % (
                value['max_size'],
                ''.join([
                    '<li><span class="hh-type">%s</span>: %s</li>' % (key, desc)
                    for key, desc in value['hh_list']
                ])
            )
        )

    @classmethod
    def get_violation_definitions(cls):
        return {
            'page.heading_hierarchy.size': {
                'title': 'Maximum size of a heading hierarchy',
                'description': cls.get_violation_description,
                'category': 'SEO'
            }
        }

    def validate(self):
        hh_list = self.review.data.get('page.heading_hierarchy', [])

        violations = []
        for key, desc in hh_list:
            if len(desc) > self.config.MAX_HEADING_HIEARARCHY_SIZE:
                violations.append((key, desc))

        if violations:
            self.add_violation(
                key='page.heading_hierarchy.size',
                value={
                    'max_size': self.config.MAX_HEADING_HIEARARCHY_SIZE,
                    'hh_list': violations
                },
                points=20 * len(violations)
            )


class H1HeadingValidator(Validator):

    @classmethod
    def get_violation_description(cls, value=None):
        return (
            'Your page does not contain any H1 headings. H1 headings help '
            'indicate the important topics of your page to search engines. '
            'While less important than good meta-titles and descriptions, '
            'H1 headings may still help define the topic of your page to '
            'search engines.'
        )

    @classmethod
    def get_violation_definitions(cls):
        return {
            'page.heading_hierarchy.h1_not_found': {
                'title': 'H1 Headings not found',
                'description': cls.get_violation_description,
                'category': 'SEO'
            }
        }

    def validate(self):
        hh_list = self.review.data.get('page.heading_hierarchy', [])
        for key, desc in hh_list:
            if key == 'h1':
                return

        self.add_violation(
            key='page.heading_hierarchy.h1_not_found',
            value=None,
            points=30
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

import sqlalchemy as sa
from datetime import datetime, timedelta

from holmes.models.worker import Worker
from holmes.handlers import BaseHandler


class WorkersHandler(BaseHandler):

    def get(self):
        zombie_time = self.application.config.ZOMBIE_WORKER_TIME
        dt = datetime.utcnow() - timedelta(seconds=zombie_time)
        workers = self.db.query(Worker).filter(Worker.last_ping > dt).all()

        workers_json = [worker.to_dict() for worker in workers]
        self.write_json(workers_json)


class WorkersInfoHandler(BaseHandler):

    def get(self):
        total_workers = self.db \
            .query(sa.func.count(Worker.id).label('count')) \
            .one()

        inactive_workers = self.db \
            .query(sa.func.count(Worker.id).label('count')) \
            .filter(Worker.current_url == None) \
            .one()

        workers_info = {
            'total': total_workers.count,
            'inactive': inactive_workers.count,
            'active': total_workers.count - inactive_workers.count,
        }
        self.write_json(workers_info)



"""Add last_review_date to page

Revision ID: 34bd7155169
Revises: 122a0b422e76
Create Date: 2013-11-27 15:16:23.976269

"""

# revision identifiers, used by Alembic.
revision = '34bd7155169'
down_revision = '122a0b422e76'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        'pages',
        sa.Column('last_review_date', sa.DateTime, nullable=True)
    )


def downgrade():
    op.drop_column('pages', 'last_review_date')



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.facters import Baser


class Validator(Baser):

    def __init__(self, reviewer):
        self.reviewer = reviewer
        self.url_buffer = set()

    @classmethod
    def get_violation_definitions(cls):
        raise NotImplementedError

    def add_violation(self, key, value, points):
        self.reviewer.add_violation(key, value, points)

    def validate(self):
        return True

    def enqueue(self, url):
        self.reviewer.enqueue(url)

    def test_url(self, url, response, broken_link_callback=None, moved_link_callback=None):
        status = response.status_code

        if status > 399:
            if broken_link_callback:
                broken_link_callback(url, response)
            return False

        if status == 302 or status == 307:
            if moved_link_callback:
                moved_link_callback(url, response)
            return False

        if response.url.rstrip('/') != url.rstrip('/'):
            return False

        return True

    def send_url(self, url, score, response):
        if self.test_url(url, response, self.broken_link_violation, self.moved_link_violation):
            self.url_buffer.add((url, score))

        if len(self.url_buffer) > self.config.MAX_ENQUEUE_BUFFER_LENGTH:
            self.flush()

    def flush(self):
        if not self.url_buffer:
            return

        self.enqueue(self.url_buffer)
        self.url_buffer = set()

    def broken_link_violation(self):
        text = 'broken_link_violation method need to be implemented by {0}'
        raise NotImplementedError(text.format(self.__class__.__name__))

    def moved_link_violation(self):
        text = 'moved_link_violation method need to be implemented by {0}'
        raise NotImplementedError(text.format(self.__class__.__name__))



#!/usr/bin/python
# -*- coding: utf-8 -*-

from gzip import GzipFile
from cStringIO import StringIO

import sqlalchemy.types as types
from sqlalchemy.ext.declarative import declarative_base
from ujson import dumps, loads

Base = declarative_base()


class JsonType(types.TypeDecorator):
    """Represents an immutable structure as a json-encoded string.

    Usage::

        JSONEncodedDict(255)

    """

    impl = types.String

    def process_bind_param(self, value, dialect):
        value = dumps(value)

        return value

    def process_result_value(self, value, dialect):
        if value:
            value = loads(value)
        else:
            value = None
        return value


class JsonTypeGzipped(types.TypeDecorator):
    impl = types.BLOB

    def process_bind_param(self, value, dialect):
        value = dumps(value)
        out = StringIO()
        with GzipFile(fileobj=out, mode="w") as f:
            f.write(value)
        return out.getvalue()

    def process_result_value(self, value, dialect):
        if value:
            gzipped = GzipFile(mode='r', fileobj=StringIO(value))
            out = loads(gzipped.read())
        else:
            out = ''
        return out

from holmes.models.domain import Domain  # NOQA
from holmes.models.page import Page  # NOQA
from holmes.models.review import Review  # NOQA
from holmes.models.fact import Fact  # NOQA
from holmes.models.violation import Violation  # NOQA
from holmes.models.keys import Key  # NOQA
from holmes.models.settings import Settings  # NOQA
from holmes.models.keys_category import KeysCategory  # NOQA
from holmes.models.request import Request  # NOQA
from holmes.models.user import User  # NOQA
from holmes.models.limiter import Limiter # NOQA



"""Requests: Create indexes

Revision ID: 22bc83869dbe
Revises: 556bb65b2a84
Create Date: 2014-01-27 14:02:34.377544

"""

# revision identifiers, used by Alembic.
revision = '22bc83869dbe'
down_revision = '556bb65b2a84'

from alembic import op


def upgrade():
    op.create_index(
        'idx_request',
        'requests',
        ['completed_date', 'domain_name', 'status_code'])


def downgrade():
    op.drop_index('idx_request', 'requests'),



"""Removing settings table

Revision ID: 2932df901655
Revises: 155c6ce689ed
Create Date: 2014-04-03 10:45:09.592592

"""

# revision identifiers, used by Alembic.
revision = '2932df901655'
down_revision = '155c6ce689ed'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.drop_table('settings')


def downgrade():
    op.create_table(
        'settings',
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column(
            'lambda_score',
            sa.Float,
            server_default=sa.text('0.0'),
            nullable=False
        )
    )

    connection = op.get_bind()
    connection.execute('INSERT INTO settings(lambda_score) VALUES(0.0)')



"""change description column

Revision ID: 122a0b422e76
Revises: 2a535a54aa95
Create Date: 2013-11-20 11:25:09.363983

"""

# revision identifiers, used by Alembic.
revision = '122a0b422e76'
down_revision = '2a535a54aa95'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.alter_column('violations', 'description', type_=sa.Text, nullable=False)


def downgrade():
    op.alter_column('violations', 'description', type_=sa.String(2000), nullable=False)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import sqlalchemy as sa

from holmes.models import Base


class Settings(Base):
    __tablename__ = "settings"

    id = sa.Column(sa.Integer, primary_key=True)
    lambda_score = sa.Column(sa.Float, default=0.0)

    @classmethod
    def instance(cls, db):
        return db.query(Settings).first()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import inspect
from uuid import uuid4
from functools import partial
from collections import defaultdict

from holmes.cli import BaseCLI
from holmes.models.domain import Domain
from holmes.models.page import Page
from holmes.models.violation import Violation
from holmes.models.request import Request
from holmes.utils import get_domain_from_url


def configure_materials(girl, db, config):
    girl.add_material(
        'domains_details',
        partial(Domain.get_domains_details, db),
        config.MATERIALS_EXPIRATION_IN_SECONDS['domains_details'],
        config.MATERIALS_GRACE_PERIOD_IN_SECONDS['domains_details']
    )

    girl.add_material(
        'next_jobs_count',
        partial(Page.get_next_jobs_count, db, config),
        config.MATERIALS_EXPIRATION_IN_SECONDS['next_jobs_count'],
        config.MATERIALS_GRACE_PERIOD_IN_SECONDS['next_jobs_count']
    )

    girl.add_material(
        'violation_count_by_category_for_domains',
        partial(Violation.get_group_by_category_id_for_all_domains, db),
        config.MATERIALS_EXPIRATION_IN_SECONDS['violation_count_by_category_for_domains'],
        config.MATERIALS_GRACE_PERIOD_IN_SECONDS['violation_count_by_category_for_domains']
    )

    girl.add_material(
        'blacklist_domain_count',
        partial(MaterialConveyor.get_blacklist_domain_count, db),
        config.MATERIALS_EXPIRATION_IN_SECONDS['blacklist_domain_count'],
        config.MATERIALS_GRACE_PERIOD_IN_SECONDS['blacklist_domain_count']
    )

    girl.add_material(
        'most_common_violations',
        partial(
            Violation.get_most_common_violations_names,
            db,
            config.get('MOST_COMMON_VIOLATIONS_SAMPLE_LIMIT')
        ),
        config.MATERIALS_EXPIRATION_IN_SECONDS['most_common_violations'],
        config.MATERIALS_GRACE_PERIOD_IN_SECONDS['most_common_violations']
    )

    girl.add_material(
        'old_requests',
        partial(Request.delete_old_requests, db, config),
        get_func_or_value(config.MATERIALS_EXPIRATION_IN_SECONDS['old_requests'], config),
        get_func_or_value(config.MATERIALS_GRACE_PERIOD_IN_SECONDS['old_requests'], config)
    )

def get_func_or_value(value, config):
    if inspect.ismethod(value) or inspect.isfunction(value):
        return value(config)

    return value


class MaterialConveyor(object):
    @classmethod
    def get_blacklist_domain_count(cls, db):
        ungrouped = defaultdict(int)
        for urls, count in Violation.get_group_by_value_for_key(db, 'blacklist.domains'):
            for url in urls:
                domain, null = get_domain_from_url(url)
                ungrouped[domain] += count
        blacklist = sorted(ungrouped.items(), key=lambda xz: -xz[1])
        return [dict(zip(('domain', 'count'), x)) for x in blacklist]


class MaterialWorker(BaseCLI):
    def initialize(self):
        self.uuid = uuid4().hex

        self.error_handlers = [handler(self.config) for handler in self.load_error_handlers()]

        self.connect_sqlalchemy()
        self.connect_to_redis()

        self.configure_material_girl()

    def do_work(self):
        self.info('Running material girl...')
        self.db.begin(subtransactions=True)
        try:
            self.girl.run()
            self.db.commit()
        except:
            self.db.rollback()
            raise


def main():
    worker = MaterialWorker(sys.argv[1:])
    worker.run()

if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

import email.utils as eut
from datetime import datetime

from holmes.facters import Facter


class LastModifiedFacter(Facter):

    @classmethod
    def get_fact_definitions(cls):
        return {
            'page.last_modified': {
                'title': 'Last-Modified',
                'description': lambda value: value,
                'category': 'HTTP'
            }
        }

    def get_facts(self):
        last_modified = self.get_last_modified()

        if not last_modified:
            return

        self.review.data['page.last_modified'] = last_modified
        self.add_fact(key='page.last_modified', value=last_modified)

    def get_last_modified(self):
        last_modified = None

        headers = self.reviewer.current.headers

        if 'Last-Modified' in headers:
            lt = headers['Last-Modified']
            last_modified = datetime(*eut.parsedate(lt)[:6])

        return last_modified



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
from uuid import uuid4
from datetime import datetime, timedelta

from ujson import dumps
from colorama import Fore, Style
from octopus import TornadoOctopus
from octopus.limiter.redis.per_domain import Limiter
from retools.lock import Lock, LockTimeout

from holmes import __version__
from holmes.reviewer import Reviewer, InvalidReviewError
from holmes.utils import load_classes, count_url_levels, get_domain_from_url
from holmes.models import Page
from holmes.cli import BaseCLI


class BaseWorker(BaseCLI):
    def _load_validators(self):
        return load_classes(default=self.config.VALIDATORS)

    def _load_facters(self):
        return load_classes(default=self.config.FACTERS)

    def get_otto_limiter(self):
        domains = self.cache.get_domain_limiters()
        limiter = None

        if domains:
            limiter = Limiter(
                *domains,
                redis=self.redis,
                expiration_in_seconds=self.config.LIMITER_LOCKS_EXPIRATION
            )

            limiter.subscribe_to_lock_miss(self.handle_limiter_miss)

        return limiter

    def update_otto_limiter(self):
        domains = self.cache.get_domain_limiters()

        if hasattr(self.otto, 'limiter') and self.otto.limiter is not None:
            self.otto.limiter.update_domain_definitions(*domains)

    def start_otto(self):
        self.info('Starting Octopus with %d concurrent threads.' % self.options.concurrency)
        self.otto = TornadoOctopus(
            concurrency=self.options.concurrency, cache=self.options.cache,
            connect_timeout_in_seconds=self.config.CONNECT_TIMEOUT_IN_SECONDS,
            request_timeout_in_seconds=self.config.REQUEST_TIMEOUT_IN_SECONDS,
            limiter=self.get_otto_limiter()
        )
        self.otto.start()

    def handle_error(self, exc_type, exc_value, tb):
        for handler in self.error_handlers:
            handler.handle_exception(
                exc_type, exc_value, tb, extra={
                    'worker-uuid': self.uuid,
                    'holmes-version': __version__
                }
            )

    def async_get(self, url, handler, method='GET', **kw):
        url, response = self.cache.get_request(url)

        if not response:
            kw['proxy_host'] = self.config.HTTP_PROXY_HOST
            kw['proxy_port'] = self.config.HTTP_PROXY_PORT

            self.debug('Enqueueing %s for %s...' % (method, url))
            self.otto.enqueue(url, self.handle_response(url, handler), method, **kw)
        else:
            handler(url, response)

    def handle_response(self, url, handler):
        def handle(url, response):
            self.cache.set_request(
                url, response.status_code, response.headers, response.cookies,
                response.text, response.effective_url, response.error, response.request_time,
                self.config.REQUEST_CACHE_EXPIRATION_IN_SECONDS
            )
            handler(url, response)
        return handle

    def handle_limiter_miss(self, url):
        pass

    def publish(self, data):
        self.redis_pub_sub.publish('events', data)

    def _insert_keys(self, keys):
        from holmes.models import Key

        for name in keys.keys():
            key = Key.get_or_create(self.db, name)
            keys[name]['key'] = key
            self.db.add(key)


class HolmesWorker(BaseWorker):
    def initialize(self):
        self.uuid = uuid4().hex
        self.working_url = None
        self.domain_name = None
        self.last_ping = None
        self.last_update_pages_score = None

        self.facters = self._load_facters()
        self.validators = self._load_validators()
        self.error_handlers = [handler(self.config) for handler in self.load_error_handlers()]

        self.connect_sqlalchemy()
        self.connect_to_redis()
        self.start_otto()

        self.fact_definitions = {}
        self.violation_definitions = {}

        for facter in self.facters:
            self.fact_definitions.update(facter.get_fact_definitions())

        self._insert_keys(self.fact_definitions)

        for validator in self.validators:
            self.violation_definitions.update(validator.get_violation_definitions())

        self._insert_keys(self.violation_definitions)

    def config_parser(self, parser):
        parser.add_argument(
            '--concurrency',
            '-t',
            type=int,
            default=10,
            help='Number of threads (or async http requests) to use for Octopus (doing GETs concurrently)'
        )

        parser.add_argument(
            '--cache', default=False, action='store_true',
            help='Whether http requests should be cached by Octopus.'
        )

    def get_description(self):
        uuid = str(getattr(self, 'uuid', ''))

        return "%s%sholmes-worker-%s%s" % (
            Fore.BLUE,
            Style.BRIGHT,
            uuid,
            Style.RESET_ALL,
        )

    def do_work(self):
        has_errored = False

        try:
            self.debug('Started doing work...')

            dt = datetime.utcnow() - timedelta(seconds=self.config.UPDATE_PAGES_SCORE_SLEEP_TIME)

            if not self.last_update_pages_score or self.last_update_pages_score < dt:
                self._update_pages_score()

            err = None
            self.update_otto_limiter()
            job = self._load_next_job()

            if job is None:
                self.info('No jobs could be found! Returning...')
                self._ping_api()
                return

            if not self._start_job(job):
                self.info('Could not start job for url "%s". Maybe other worker doing it?' % job['url'])
                lock = job.get('lock', None)
                self._release_lock(lock)
                return

            try:
                self.info('Starting new job for %s...' % job['url'])
                self._start_reviewer(job=job)
            except InvalidReviewError:
                err = str(sys.exc_info()[1])
                self.error("Fail to review %s: %s" % (job['url'], err))
                raise

            lock = job.get('lock', None)
            self._complete_job(lock)

        except Exception:
            has_errored = True
            err = str(sys.exc_info()[1])
            self.error("Fail to complete work: %s" % err)
            self.db.rollback()
            raise
        finally:
            if not has_errored:
                self.db.commit()

    def _start_reviewer(self, job):
        if job:

            if count_url_levels(job['url']) > self.config.MAX_URL_LEVELS:
                self.info('Max URL levels! Details: %s' % job['url'])
                return

            self.debug('Starting Review for [%s]' % job['url'])
            reviewer = Reviewer(
                api_url=self.config.HOLMES_API_URL,
                page_uuid=job['page'],
                page_url=job['url'],
                page_score=0,
                config=self.config,
                validators=self.validators,
                facters=self.facters,
                async_get=self.async_get,
                wait=self.otto.wait,
                wait_timeout=0,  # max time to wait for all requests to finish
                db=self.db,
                cache=self.cache,
                publish=self.publish,
                fact_definitions=self.fact_definitions,
                violation_definitions=self.violation_definitions
            )

            reviewer.review()

    def _ping_api(self):
        self.debug('Pinging that this worker is still alive...')

        self.last_ping = datetime.utcnow()

        self.publish(dumps({
            'type': 'worker-status',
            'workerId': str(self.uuid),
            'dt': self.last_ping,
            'url': self.working_url,
            'domainName': self.domain_name,
        }))

    def handle_limiter_miss(self, url):
        self.working_url = url

        if self.last_ping < datetime.utcnow() - timedelta(seconds=1):
            self._ping_api()

    def _load_next_job(self):
        return self.cache.get_next_job(self.config.REVIEW_EXPIRATION_IN_SECONDS, self.config.WORKERS_LOOK_AHEAD_PAGES)

    def _start_job(self, job):
        try:
            lock = Lock(job['url'], redis=self.redis, timeout=1)
            lock.acquire()

            self.working_url = job['url']

            if self.working_url:
                self.domain_name, _ = get_domain_from_url(self.working_url)

            self._ping_api()
            job['lock'] = lock

            return True
        except LockTimeout:
            job['lock'] = None
            return False

    def _complete_job(self, lock):
        self.working_url = None
        self.domain_name = None
        self._ping_api()
        self._release_lock(lock)

    def _release_lock(self, lock):
        if lock is not None:
            lock.release()

    def _update_pages_score(self):
        expiration = self.config.UPDATE_PAGES_SCORE_EXPIRATION
        lock = self.cache.has_update_pages_lock(expiration)

        if lock is not None:
            self.debug('Updating pages score...')
            Page.update_pages_score(self.db, self.cache, self.config)
            self.cache.release_update_pages_lock(lock)
            self.last_update_pages_score = datetime.utcnow()


def main():
    worker = HolmesWorker(sys.argv[1:])
    worker.run()

if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging

from holmes.facters import Facter


class ImageFacter(Facter):
    @classmethod
    def get_fact_definitions(cls):
        return {
            'page.images': {
                'title': 'Images',
                'description': lambda value: list(value),
                'unit': 'images',
                'category': 'Static',
            },
            'total.size.img': {
                'title': 'Total images size',
                'description': lambda value: '%d' % value,
                'unit': 'kb',
                'category': 'SEO',
            },
            'total.requests.img': {
                'title': 'Total images requests',
                'description': lambda value: '%d' % value,
                'category': 'HTTP',
            }
        }

    def looks_like_base64(self, url):
        if url.lstrip().startswith('data:image'):
            return True
        return False

    def get_facts(self):
        img_files = self.get_images()

        self.review.data['page.images'] = set()
        self.review.data['total.size.img'] = 0

        self.add_fact(
            key='page.images',
            value=set()
        )

        self.add_fact(
            key='total.size.img',
            value=0
        )

        images_to_get = set()
        images_without_base64 = []

        for img_file in img_files:
            src = img_file.get('src')
            if not src:
                continue

            if self.looks_like_base64(src):
                continue

            images_without_base64.append(img_file)

            src = self.normalize_url(src)
            if src:
                images_to_get.add(src)

        self.review.data['page.all_images'] = images_without_base64

        for src in images_to_get:
            self.async_get(src, self.handle_url_loaded)

        self.add_fact(
            key='total.requests.img',
            value=len(images_to_get)
        )

    def handle_url_loaded(self, url, response):
        logging.debug('Got response (%s) from %s!' % (response.status_code,
                                                      url))

        size_img = 0
        if response.text:
            size_img = len(response.text) / 1024.0

        self.review.facts['page.images']['value'].add(url)
        self.review.data['page.images'].add((url, response))

        self.review.facts['total.size.img']['value'] += size_img
        self.review.data['total.size.img'] += size_img

    def get_images(self):
        return self.reviewer.current_html.cssselect(':not(script) img[src]')



"""Refactoring get_active_reviews

Revision ID: 54ca0bd9bc1e
Revises: cc4091b07cb
Create Date: 2014-01-03 15:58:28.836246

"""

# revision identifiers, used by Alembic.
revision = '54ca0bd9bc1e'
down_revision = 'cc4091b07cb'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('pages', sa.Column('violations_count', sa.Integer, server_default='0', nullable=False))
    op.create_index('idx_violations_count', 'pages', ['violations_count'])

    op.add_column('pages', sa.Column('last_review_uuid', sa.String(36), nullable=True))
    op.create_index('idx_last_review_uuid', 'pages', ['last_review_uuid'])

    connection = op.get_bind()
    query = (
        'SELECT count(violations.id) AS violations_count, reviews.uuid, reviews.page_id '
        'FROM reviews LEFT OUTER JOIN violations ON violations.review_id = reviews.id '
        'WHERE reviews.is_active = 1 '
        'GROUP BY reviews.id'
    )

    for values in connection.execute(query):
        update = (
            'UPDATE pages '
            'SET violations_count = \'{0}\', last_review_uuid = \'{1}\' '
            'WHERE `id` = \'{2}\''.format(*values)
        )
        connection.execute(update)


def downgrade():
    op.drop_index('idx_violations_count', 'pages')
    op.drop_column('pages', 'violations_count')

    op.drop_index('idx_last_review_uuid', 'pages')
    op.drop_column('pages', 'last_review_uuid')



"""Renaming delimiters table to limiters

Revision ID: 17346cf564bc
Revises: 4d45dd3d8ce5
Create Date: 2014-03-07 14:45:27.909631

"""

# revision identifiers, used by Alembic.
revision = '17346cf564bc'
down_revision = '4d45dd3d8ce5'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.rename_table('delimiters', 'limiters')


def downgrade():
    op.rename_table('limiters', 'delimiters')



#!/usr/bin/python
# -*- coding: utf-8 -*-

import urlparse
from holmes.utils import is_valid


class Baser(object):

    def __init__(self, reviewer):
        self.reviewer = reviewer

    @property
    def page_uuid(self):
        return self.reviewer.page_uuid

    @property
    def page_url(self):
        return self.reviewer.page_url

    @property
    def review(self):
        return self.reviewer.review_dao

    @property
    def config(self):
        return self.reviewer.config

    def is_absolute(self, url):
        return bool(urlparse.urlparse(url).scheme)

    def rebase(self, url):
        return urlparse.urljoin(self.page_url, url)

    def normalize_url(self, url):
        parse = is_valid(url)
        if parse:
            if not self.is_absolute(url):
                url = self.rebase(url)
            return url
        return None

    def to_gzip(self, content):
        return content.encode('zip')

    def add_fact(self, key, value):
        self.reviewer.add_fact(key, value)


class Facter(Baser):

    unit = 'value'

    @classmethod
    def get_fact_definitions(cls):
        raise NotImplementedError

    def add_violation(self, key, value):
        self.reviewer.add_violation(key, value)

    def get(self, url):
        return self.reviewer._get(url)

    def async_get(self, url, handler, method='GET', **kw):
        self.reviewer._async_get(url, handler, method, **kw)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.validators.base import Validator


class CSSRequestsValidator(Validator):
    @classmethod
    def get_requests_css_message(cls, value):
        return 'This page has %d CSS request (%d over limit). ' \
               'Having too many requests impose a tax in the ' \
               'browser due to handshakes.' % (value['total_css_files'],
                                               value['over_limit'])

    @classmethod
    def get_total_size_message(cls, value):
        return 'There\'s %.2fkb of CSS in this page and that ' \
               'adds up to more download time slowing down the ' \
               'page rendering to the user.' % value

    @classmethod
    def get_violation_definitions(cls):
        return {
            'total.requests.css': {
                'title': 'Too many CSS requests',
                'description': cls.get_requests_css_message,
                'category': 'Performance'
            },
            'total.size.css': {
                'title': 'CSS size in kb is too big',
                'description': cls.get_total_size_message,
                'category': 'Performance'
            },
        }

    def validate(self):
        total_css_files = self.get_total_requests_css()
        total_size_gzip = self.get_total_size_css_gzipped()

        max_requests = self.reviewer.config.MAX_CSS_REQUESTS_PER_PAGE
        over_limit = total_css_files - max_requests

        if total_css_files > self.reviewer.config.MAX_CSS_REQUESTS_PER_PAGE:
            self.add_violation(
                key='total.requests.css',
                value={
                    'total_css_files': total_css_files,
                    'over_limit': over_limit
                },
                points=5 * over_limit
            )

        max_kb_gzip = self.reviewer.config.MAX_CSS_KB_PER_PAGE_AFTER_GZIP

        if total_size_gzip > max_kb_gzip:
            self.add_violation(
                key='total.size.css',
                value=total_size_gzip,
                points=int(total_size_gzip - max_kb_gzip)
            )

    def get_css_requests(self):
        return self.review.data.get('page.css', None)

    def get_total_requests_css(self):
        return self.review.data.get('total.requests.css', 0)

    def get_total_size_css_gzipped(self):
        return self.review.data.get('total.size.css.gzipped', 0)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging

from holmes.facters import Facter


class RobotsFacter(Facter):

    @classmethod
    def get_fact_definitions(cls):
        return {
            'robots.url': {
                'title': 'Robots file URL',
                'description': lambda value: value,
                'category': 'SEO'
            }
        }

    def get_facts(self):
        if not self.reviewer.is_root():
            return

        url = self.rebase('/robots.txt')

        self.review.data['robots.response'] = None

        self.async_get(url, self.handle_robots_loaded)

    def handle_robots_loaded(self, url, response):
        logging.debug('Got response (%s) from %s!' % (response.status_code,
                                                      url))

        if url == self.rebase('/robots.txt') and response.status_code <= 399:
            self.add_fact(
                key='robots.url',
                value=url
            )

        self.review.data['robots.response'] = response



#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging

from holmes.facters import Facter


class CSSFacter(Facter):
    @classmethod
    def get_fact_definitions(cls):
        return {
            'page.css': {
                'title': 'CSS',
                'description': lambda value: list(value),
                'unit': 'css',
                'category': 'Static',
            },
            'total.requests.css': {
                'title': 'Total CSS requests',
                'description': lambda value: '%d' % value,
                'category': 'HTTP',
            },
            'total.size.css': {
                'title': 'Total CSS size',
                'description': lambda value: '%d' % value,
                'unit': 'kb',
                'category': 'SEO',
            },
            'total.size.css.gzipped': {
                'title': 'Total CSS size gzipped',
                'description': lambda value: '%d' % value,
                'unit': 'kb',
                'category': 'SEO',
            }
        }

    def get_facts(self):
        css_files = self.get_css()

        self.review.data['page.css'] = set()
        self.review.data['total.size.css'] = 0
        self.review.data['total.size.css.gzipped'] = 0

        self.add_fact(
            key='page.css',
            value=set(),
        )

        self.add_fact(
            key='total.size.css',
            value=0,
        )

        self.add_fact(
            key='total.size.css.gzipped',
            value=0,
        )

        num_css = 0

        css_to_get = []

        for css_file in css_files:
            src = css_file.get('href')
            if not src.endswith('.css'):
                continue

            src = self.normalize_url(src)
            if src:
                css_to_get.append(src)
                num_css += 1

        self.add_fact(
            key='total.requests.css',
            value=num_css,
        )

        for url in css_to_get:
            self.async_get(url, self.handle_url_loaded)

    def handle_url_loaded(self, url, response):
        logging.debug('Got response (%s) from %s!' % (response.status_code,
                                                      url))

        self.review.facts['page.css']['value'].add(url)
        self.review.data['page.css'].add((url, response))

        if response.text:
            size_css = len(response.text) / 1024.0
            size_gzip = len(self.to_gzip(response.text)) / 1024.0
        else:
            size_css = 0
            size_gzip = 0

        self.review.facts['total.size.css']['value'] += size_css
        self.review.data['total.size.css'] += size_css

        self.review.facts['total.size.css.gzipped']['value'] += size_gzip
        self.review.data['total.size.css.gzipped'] += size_gzip

    def get_css(self):
        return self.reviewer.current_html.cssselect('link[href]')



"""create review table

Revision ID: 5048316f2933
Revises: 2cb79bf80b46
Create Date: 2013-11-13 10:39:50.428863

"""

# revision identifiers, used by Alembic.
revision = '5048316f2933'
down_revision = '2cb79bf80b46'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'reviews',
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('is_active', sa.Boolean, default=False, nullable=False),
        sa.Column('is_complete', sa.Boolean, default=False, nullable=False),
        sa.Column('uuid', sa.String(36), nullable=False),
        sa.Column('created_date', sa.DateTime, nullable=False),
        sa.Column('completed_date', sa.DateTime, nullable=True),
        sa.Column('last_review_started_date', sa.DateTime, nullable=True),
        sa.Column('last_review_started_date', sa.DateTime, nullable=True),
        sa.Column('failure_message', sa.String(2000), nullable=True),
        sa.Column('domain_id', sa.Integer),
        sa.Column('page_id', sa.Integer),
    )

    op.create_foreign_key(
        "fk_review_domain", "reviews",
        "domains", ["domain_id"], ["id"]
    )

    op.create_foreign_key(
        "fk_review_page", "reviews",
        "pages", ["page_id"], ["id"]
    )


def downgrade():
    op.drop_constraint('fk_review_domain', 'reviews', type_="foreignkey")
    op.drop_constraint('fk_review_page', 'reviews', type_="foreignkey")
    op.drop_table('reviews')



#!/usr/bin/python
# -*- coding: utf-8 -*-

import datetime
from tornado.gen import coroutine

from holmes.utils import get_status_code_title
from holmes.models import Request
from holmes.handlers import BaseHandler


class RequestDomainHandler(BaseHandler):

    @coroutine
    def get(self, domain_name, status_code):

        if not domain_name:
            self.set_status(404, 'Domain %s not found' % domain_name)
            return

        if not status_code:
            self.set_status(404, 'Status code %s not found' % status_code)
            return

        if status_code == '200':
            self.set_status(403, 'Status code %s is not allowed' % status_code)
            return

        current_page = int(self.get_argument('current_page', 1))
        page_size = int(self.get_argument('page_size', 10))

        requests = Request.get_requests_by_status_code(
            domain_name,
            status_code,
            self.db,
            current_page=current_page,
            page_size=page_size
        )

        requests_count = Request.get_requests_by_status_count(
            domain_name,
            status_code,
            self.db
        )

        result = {
            'statusCodeTitle': get_status_code_title(status_code),
            'requestsCount': requests_count,
            'requests': []
        }

        for request in requests:
            result['requests'].append({
                'id': request.id,
                'url': request.url,
                'review_url': request.review_url,
                'completed_date': request.completed_date
            })

        self.write_json(result)


class LastRequestsHandler(BaseHandler):
    @coroutine
    def get(self):
        current_page = int(self.get_argument('current_page', 1))
        page_size = int(self.get_argument('page_size', 10))

        requests = Request.get_last_requests(
            self.db,
            current_page=current_page,
            page_size=page_size
        )

        requests_count = yield self.cache.get_requests_count()

        result = {'requestsCount': requests_count, 'requests': []}

        for request in requests:
            result['requests'].append(request.to_dict())

        self.write_json(result)


class RequestsInLastDayHandler(BaseHandler):
    @coroutine
    def get(self):
        from_date = datetime.datetime.utcnow() - datetime.timedelta(days=1)
        requests = Request.get_requests_count_by_status_in_period_of_days(self.db, from_date=from_date)

        result = []
        for request in requests:
            result.append({
                'statusCode': request.status_code,
                'statusCodeTitle': get_status_code_title(request.status_code),  # FIXME: is it code or title??
                'count': request.count,
            })

        self.write_json(result)



#!/usr/bin/python
# -*- coding: utf-8 -*-


from holmes.validators.base import Validator


class RobotsValidator(Validator):

    SITEMAP_NOT_FOUND = 'You must specify the location of the Sitemap ' \
                        'using a robots.txt file'

    DISALLOW_NOT_FOUND = 'Disallow directive indicates that robots should ' \
                         'not access the specific directory, subdirectory ' \
                         'or file. '

    DISALLOW_ROOT_PATH = 'Crawlers may not index anything, because ' \
                         'the root path (/) is disallowed.'

    @classmethod
    def get_violation_definitions(cls):
        return {
            'robots.not_found': {
                'title': 'Robots file not found',
                'description': lambda value: "The robots file at '%s' was not found." % value,
                'category': 'SEO'
            },
            'robots.empty': {
                'title': 'Robots file was empty',
                'description': lambda value: "The robots file at '%s' was empty." % value,
                'category': 'SEO'
            },
            'robots.sitemap.not_found': {
                'title': 'Sitemap in Robots not found',
                'description': lambda value: cls.SITEMAP_NOT_FOUND,
                'category': 'SEO'
            },
            'robots.disallow.not_found': {
                'title': 'Disallow in Robots not found',
                'description': lambda value: cls.DISALLOW_NOT_FOUND,
                'category': 'SEO'
            },
            'robots.disallow.root_path': {
                'title': 'Disallow: / in Robots',
                'description': lambda value: cls.DISALLOW_ROOT_PATH,
                'category': 'SEO'
            },
        }

    def validate(self):
        if not self.reviewer.is_root():
            return

        response = self.review.data['robots.response']

        if response.status_code > 399:
            self.add_violation(
                key='robots.not_found',
                value=response.url,
                points=100
            )
            return

        if not response.text.strip():
            self.add_violation(
                key='robots.empty',
                value=response.url,
                points=100
            )
            return

        has_sitemap = False

        has_disallow = False
        disallow_root_path = False

        for rawline in response.text.splitlines():
            line = rawline.strip()
            comments = line.find('#')
            if comments >= 0:
                line = line[:comments]
            if line == '' or ':' not in line:
                continue
            key, val = [x.strip() for x in line.split(':', 1)]
            key = key.lower()
            if key == 'sitemap':
                has_sitemap = True
            elif key == 'disallow':
                has_disallow = True
                if val == '/':
                    disallow_root_path = True

        if not has_sitemap:
            self.add_violation(
                key='robots.sitemap.not_found',
                value=None,
                points=100
            )

        if not has_disallow:
            self.add_violation(
                key='robots.disallow.not_found',
                value=None,
                points=100
            )
        elif disallow_root_path:
            self.add_violation(
                key='robots.disallow.root_path',
                value=None,
                points=100
            )



#!/usr/bin/python
# -*- coding: utf-8 -*-

import sqlalchemy as sa
from sqlalchemy.orm import relationship

from holmes.models import Base


class KeysCategory(Base):
    __tablename__ = "keys_category"

    id = sa.Column(sa.Integer, primary_key=True)
    name = sa.Column('name', sa.String(255), nullable=False)

    keys = relationship("Key", cascade="all,delete", backref="category")

    def __str__(self):
        return '%s' % (self.name)

    def __repr__(self):
        return str(self)

    def to_dict(self):
        return {
            'name': self.name
        }

    @classmethod
    def get_or_create(cls, db, category_name):
        category = db \
            .query(KeysCategory) \
            .filter(KeysCategory.name == category_name) \
            .scalar()

        if not category:
            category = KeysCategory(name=category_name)

        return category



"""Wrong index for last_review_date

Revision ID: 49a08fc2113a
Revises: 3a3aa7483604
Create Date: 2013-11-27 17:53:25.879893

"""

# revision identifiers, used by Alembic.
revision = '49a08fc2113a'
down_revision = '3a3aa7483604'

from alembic import op


def upgrade():
    op.drop_index('idx_pages_last_review_date', 'pages')
    op.create_index('idx_pages_last_review_date', 'pages', ['last_review_date'])


def downgrade():
    op.drop_index('idx_pages_last_review_date', 'pages')
    op.create_index('idx_pages_last_review_date', 'pages', ['last_review_started_date'])



"""Violation: value can be null

Revision ID: 2f69133b476d
Revises: 1e071953a65d
Create Date: 2013-12-06 16:16:49.228679

"""

# revision identifiers, used by Alembic.
revision = '2f69133b476d'
down_revision = '1e071953a65d'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.alter_column('violations', 'value',  type_=sa.Text, nullable=True)


def downgrade():
    op.alter_column('violations', 'value',  type_=sa.Text, nullable=False)



"""add a flag for review active

Revision ID: 4d45dd3d8ce5
Revises: 49d09b3d2801
Create Date: 2014-02-26 16:29:54.507710

"""

# revision identifiers, used by Alembic.
revision = '4d45dd3d8ce5'
down_revision = '49d09b3d2801'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        'violations',
        sa.Column('review_is_active', sa.Integer, nullable=False, server_default='1')
    )

    connection = op.get_bind()
    connection.execute('''UPDATE violations
        SET review_is_active = 0
        WHERE review_id IN (SELECT id FROM reviews WHERE is_active = 0)''')

    op.create_index(
        'idx_key_domain_review_active',
        'violations',
        ['key_id', 'domain_id', 'review_is_active'])


def downgrade():
    op.drop_index('idx_key_domain_review_active', 'violations'),
    op.drop_column('violations', 'review_is_active')



#!/usr/bin/python
# -*- coding: utf-8 -*-

import datetime
from uuid import UUID

from holmes.models import Review, Page
from holmes.handlers import BaseHandler


class BaseReviewHandler(BaseHandler):
    def _parse_uuid(self, uuid):
        try:
            return UUID(uuid)
        except ValueError:
            return None


class ReviewHandler(BaseReviewHandler):
    def get(self, page_uuid, review_uuid):
        review = None
        page = None
        if self._parse_uuid(review_uuid):
            review = Review.by_uuid(review_uuid, self.db)

        if self._parse_uuid(page_uuid):
            page = Page.by_uuid(page_uuid, self.db)

        if not review and page:
            self.redirect('/page/%s/review/%s/' % (page_uuid, page.last_review_uuid))
            return

        if not page:
            self.set_status(404, 'Page with uuid of %s not found!' % page_uuid)
            return

        result = review.to_dict(self.application.fact_definitions, self.application.violation_definitions)
        result.update({
            'violationPoints': review.get_violation_points(),
            'violationCount': review.violation_count,
        })

        self.write_json(result)


class LastReviewsHandler(BaseReviewHandler):
    def get(self):
        reviews = Review.get_last_reviews(self.db)

        reviews_json = []
        for review in reviews:
            review_dict = review.to_dict(self.application.fact_definitions, self.application.violation_definitions)
            data = {
                'violationCount': review.violation_count,
            }
            review_dict.update(data)
            reviews_json.append(review_dict)

        self.write_json(reviews_json)


class ReviewsInLastHourHandler(BaseReviewHandler):
    def get(self):
        from_date = datetime.datetime.utcnow() - datetime.timedelta(hours=1)
        count, first_date = Review.get_reviews_count_in_period(
            self.db,
            from_date=from_date
        )

        if first_date:
            ellapsed = (datetime.datetime.utcnow() - first_date).total_seconds()
        else:
            ellapsed = 3600

        self.write_json({'count': count, 'ellapsed': ellapsed})



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.validators.base import Validator


class ImageAltValidator(Validator):
    @classmethod
    def get_without_alt_message(cls, value):
        result = []
        for src, name in value:
            data = '<a href="%s" target="_blank">%s</a>' % (src, name)
            result.append(data)

        return (
            'Images without alt text are not good for '
            'Search Engines. Images without alt were '
            'found for: %s.' % (', '.join(result))
        )

    @classmethod
    def get_alt_too_big_message(cls, value):
        result = []
        for src, name, alt in value['images']:
            data = u'<a href="{}" alt="{}" target="_blank">{}</a>'.format(
                src, alt, name
            )
            result.append(data)

        return (
            'Images with alt text bigger than %d chars are not good for '
            'Search Engines. Images with a too big alt were '
            'found for: %s.' % (value['max_size'], ', '.join(result))
        )

    @classmethod
    def get_violation_definitions(cls):
        return {
            'invalid.images.alt': {
                'title': 'Image(s) without alt attribute',
                'description': cls.get_without_alt_message,
                'category': 'SEO'
            },
            'invalid.images.alt_too_big': {
                'title': 'Image(s) with alt attribute too big',
                'description': cls.get_alt_too_big_message,
                'category': 'SEO'
            }
        }

    def validate(self):
        imgs = self.get_imgs()

        result_no_alt = []
        result_alt_too_big = []
        for img in imgs:
            src = img.get('src')
            if not src:
                continue

            src = self.normalize_url(src)
            img_alt = img.get('alt')

            if src:
                name = src.rsplit('/', 1)[-1]
                if not img_alt:
                    result_no_alt.append((src, name))
                elif len(img_alt) > self.config.MAX_IMAGE_ALT_SIZE:
                    result_alt_too_big.append((src, name, img_alt))

        if result_no_alt:
            self.add_violation(
                key='invalid.images.alt',
                value=result_no_alt,
                points=20 * len(result_no_alt)
            )

        if result_alt_too_big:
            self.add_violation(
                key='invalid.images.alt_too_big',
                value={
                    'images': result_alt_too_big,
                    'max_size': self.config.MAX_IMAGE_ALT_SIZE
                },
                points=20 * len(result_alt_too_big)
            )

    def get_imgs(self):
        return self.review.data.get('page.all_images', None)



"""idx requests by status, completed

Revision ID: 7f4a3b8c55d
Revises: 4b96dd9974bb
Create Date: 2014-03-31 17:39:40.858182

"""

# revision identifiers, used by Alembic.
revision = '7f4a3b8c55d'
down_revision = '4b96dd9974bb'

from alembic import op


def upgrade():
    op.create_index('idx_status_complete', 'requests', ['status_code', 'completed_date'])


def downgrade():
    op.drop_index('idx_status_complete', 'requests')



#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging
import re
import lxml.etree

from holmes.facters import Facter


ROBOTS_SITEMAP = re.compile('Sitemap:\s+(.*)')


class SitemapFacter(Facter):

    @classmethod
    def get_fact_definitions(cls):
        return {
            'total.sitemap.indexes': {
                'title': 'Total SiteMap indexes',
                'category': 'SEO',
            },
            'total.sitemap.urls': {
                'title': 'Total SiteMap urls',
                'category': 'SEO',
            },
            'total.size.sitemap': {
                'title': 'Total SiteMap size',
                'unit': 'kb',
                'category': 'SEO',
            },
            'total.size.sitemap.gzipped': {
                'title': 'Total SiteMap gzipped size',
                'unit': 'kb',
                'category': 'SEO',
            }
        }

    def get_facts(self):
        if not self.reviewer.is_root():
            return

        self.review.data['sitemap.data'] = {}
        self.review.data['sitemap.urls'] = {}
        self.review.data['sitemap.files'] = set()
        self.review.data['sitemap.files.size'] = {}
        self.review.data['sitemap.files.urls'] = {}
        self.review.data['total.size.sitemap'] = 0
        self.review.data['total.size.sitemap.gzipped'] = 0

        self.add_fact(
            key='total.sitemap.indexes',
            value=0
        )

        self.add_fact(
            key='total.sitemap.urls',
            value=0
        )

        self.add_fact(
            key='total.size.sitemap',
            value=0
        )

        self.add_fact(
            key='total.size.sitemap.gzipped',
            value=0
        )

        self.async_get(self.rebase('/robots.txt'), self.handle_robots_loaded)

    def get_sitemaps(self, response):
        sitemaps = set([self.rebase('/sitemap.xml')])

        if response.status_code > 399:
            return sitemaps

        sitemaps.update(ROBOTS_SITEMAP.findall(response.text))

        return sitemaps

    def handle_sitemap_loaded(self, url, response):
        logging.debug('Got sitemap %s with status %s' % (url, response.status_code))
        self.review.data['sitemap.data'][url] = response

        namespaces = [
            ('sm', 'http://www.sitemaps.org/schemas/sitemap/0.9'),
        ]

        if response.status_code > 399 or response.text is None or not response.text.strip():
            return

        self.review.facts['total.sitemap.indexes']['value'] += 1

        size_sitemap = len(response.text) / 1024.0
        size_gzip = len(self.to_gzip(response.text)) / 1024.0

        self.review.data['sitemap.files.urls'][url] = 0
        self.review.data['sitemap.files.size'][url] = size_sitemap
        self.review.data['sitemap.urls'][url] = set()

        self.review.facts['total.size.sitemap']['value'] += size_sitemap
        self.review.data['total.size.sitemap'] += size_sitemap

        self.review.facts['total.size.sitemap.gzipped']['value'] += size_gzip
        self.review.data['total.size.sitemap.gzipped'] += size_gzip

        tree = lxml.etree.fromstring(response.text)

        for sitemap in tree.xpath('//sm:sitemap | //sitemap', namespaces=namespaces):
            for loc in sitemap.xpath('sm:loc | loc', namespaces=namespaces):
                loc = loc.text.strip()
                self.review.data['sitemap.files'].add(loc)
                self.review.data['sitemap.files.urls'][url] += 1
                self.async_get(loc, self.handle_sitemap_loaded)

        for sitemap in tree.xpath('//sm:url | //url', namespaces=namespaces):
            for loc in sitemap.xpath('sm:loc | loc', namespaces=namespaces):
                loc = loc.text.strip()
                self.review.data['sitemap.urls'][url].add(loc)
                self.review.data['sitemap.files.urls'][url] += 1
                self.review.facts['total.sitemap.urls']['value'] += 1

    def handle_robots_loaded(self, url, response):
        sitemaps = self.get_sitemaps(response)

        for sitemap in sitemaps:
            self.async_get(sitemap, self.handle_sitemap_loaded)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging

from ujson import dumps
from tornado.web import RequestHandler

from holmes import __version__


class BaseHandler(RequestHandler):
    def initialize(self, *args, **kw):
        super(BaseHandler, self).initialize(*args, **kw)
        self._session = None

    def log_exception(self, typ, value, tb):
        for handler in self.application.error_handlers:
            handler.handle_exception(
                typ, value, tb, extra={
                    'url': self.request.full_url(),
                    'ip': self.request.remote_ip,
                    'holmes-version': __version__
                }
            )

        super(BaseHandler, self).log_exception(typ, value, tb)

    def on_finish(self):
        try:
            if self.application.config.COMMIT_ON_REQUEST_END:
                if self.get_status() > 399:
                    logging.debug('ROLLING BACK TRANSACTION')
                    self.db.rollback()
                else:
                    logging.debug('COMMITTING TRANSACTION')
                    self.db.flush()
                    self.db.commit()
                    self.application.event_bus.flush()
        finally:
            self.db.close()

    def options(self):
        self.set_header('Access-Control-Allow-Origin', self.application.config.ORIGIN)
        self.set_header('Access-Control-Allow-Methods', 'GET,PUT,POST,DELETE,OPTIONS')
        self.set_header('Access-Control-Allow-Headers', 'Accept, Content-Type, X-AUTH-HOLMES')
        self.set_status(200)
        self.finish()

    def set_default_headers(self):
        self.set_header('Access-Control-Allow-Origin', self.application.config.ORIGIN)
        self.set_header('Access-Control-Allow-Methods', 'GET,PUT,POST,DELETE,OPTIONS')

    def write_json(self, obj):
        self.set_header("Content-Type", "application/json")
        self.write(dumps(obj))

    @property
    def cache(self):
        return self.application.cache

    @property
    def db(self):
        #return self.application.db
        if self._session is None:
            self._session = self.application.get_sqlalchemy_session()

        return self._session

    @property
    def girl(self):
        return self.application.girl




"""Create new score index

Revision ID: 531d795ea74
Revises: 3412b4b8b24e
Create Date: 2014-01-16 16:51:54.447201

"""

# revision identifiers, used by Alembic.
revision = '531d795ea74'
down_revision = '3412b4b8b24e'

from alembic import op


def upgrade():
    op.drop_index('idx_pages_score', 'pages')

    connection = op.get_bind()
    connection.execute('CREATE index idx_last_review_date_score on pages(last_review_date, score DESC)')


def downgrade():
    op.drop_index('idx_last_review_date_score', 'pages')
    op.create_index('idx_pages_score', 'pages', ['score'])



#!/usr/bin/python
# -*- coding: utf-8 -*-

import hashlib
import math
from collections import defaultdict

import sqlalchemy as sa

from holmes.models import Base


class Limiter(Base):
    __tablename__ = "limiters"

    id = sa.Column(sa.Integer, primary_key=True)
    url = sa.Column('url', sa.String(2000), nullable=False)
    url_hash = sa.Column('url_hash', sa.String(128), nullable=False)
    value = sa.Column('value', sa.Integer, server_default='1', nullable=False)

    def to_dict(self):
        return {
            'url': self.url,
            'value': self.value
        }

    def __str__(self):
        return str(self.url)

    def __repr__(self):
        return str(self)

    @classmethod
    def get_all(cls, db):
        return db \
            .query(Limiter) \
            .order_by(sa.func.char_length(Limiter.url)) \
            .all()

    @classmethod
    def by_url(cls, url, db):
        return db.query(Limiter).filter(Limiter.url == url).first()

    @classmethod
    def by_url_hash(cls, url_hash, db):
        return db.query(Limiter).filter(Limiter.url_hash == url_hash).first()

    @classmethod
    def by_id(cls, id_, db):
        return db.query(Limiter).filter(Limiter.id == id_).first()

    @classmethod
    def delete(cls, id_, db):
        return db.query(Limiter).filter(Limiter.id == id_).delete()

    def matches(self, url):
        return url.startswith(self.url)

    @classmethod
    def add_or_update_limiter(cls, db, url, value):
        if not url:
            return

        url = url.encode('utf-8')
        url_hash = hashlib.sha512(url).hexdigest()
        limiter = Limiter.by_url_hash(url_hash, db)

        if limiter:
            db \
                .query(Limiter) \
                .filter(Limiter.id == limiter.id) \
                .update({'value': value})

            db.flush()
            db.commit()

            return limiter.url

        limiter = Limiter(url=url, url_hash=url_hash, value=value)
        db.add(limiter)

        return limiter.url

    @classmethod
    def get_limiters_for_domains(cls, db, active_domains):
        from holmes.models import Limiter  # Avoid circular dependency

        all_limiters = Limiter.get_all(db)

        limiters = []
        for limiter in all_limiters:
            for domain in active_domains:
                if limiter.matches(domain.url):
                    limiters.append(limiter)

        return limiters

    @classmethod
    def get_limiters_per_domains(cls, db, active_domains):
        from holmes.models import Limiter  # Avoid circular dependency

        all_limiters = Limiter.get_all(db)

        limiters = defaultdict(list)
        for limiter in all_limiters:
            for domain in active_domains:
                if limiter.matches(domain.url):
                    limiters[domain.id].append(limiter)

        return limiters

    @classmethod
    def _get_limiter_for_url(cls, limiters, url):
        for limiter in limiters:
            if limiter.matches(url):
                return limiter

        return None

    @classmethod
    def has_limit_to_work(cls, db, cache, active_domains, url, avg_links_per_page=10):
        if avg_links_per_page < 1:
            avg_links_per_page = 1

        limiters = Limiter.get_limiters_for_domains(db, active_domains)

        limiter = Limiter._get_limiter_for_url(limiters, url)

        if limiter:
            worker_count = cache.get_limit_usage(limiter.url)

            if worker_count >= math.ceil(float(limiter.value) / float(avg_links_per_page)):
                return False

        return True


    @classmethod
    def get_limit_capacity(cls, db, active_domains):
        limiters = Limiter.get_limiters_per_domains(db, active_domains)

        usage = defaultdict(int)

        for domain_id, domain_limiters in limiters.items():
            for limiter in domain_limiters:
                usage[domain_id] += limiter.value

        return usage



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.facters import Facter


class TitleFacter(Facter):

    @classmethod
    def get_fact_definitions(cls):
        return {
            'page.title': {
                'title': 'Page Title',
                'description': lambda value: value,
                'category': 'SEO',
            }
        }

    def get_facts(self):
        titles = self.reviewer.current_html.cssselect('title')

        if not titles:
            return

        self.review.data['page.title_count'] = len(titles)

        self.add_fact(key='page.title', value=titles[0].text)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import re
import logging

from holmes.facters import Facter
from holmes.utils import get_domain_from_url, count_url_levels

REMOVE_HASH = re.compile('([#].*)$')
URL_RE = re.compile(
    r'^(?:http|ftp)s?://'  # http:// or https://
    r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
    r'localhost|'  # localhost...
    r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
    r'(?::\d+)?'  # optional port
    r'(?:/?|[/?]\S+)$', re.IGNORECASE)


class LinkFacter(Facter):
    @classmethod
    def get_fact_definitions(cls):
        return {
            'page.links': {
                'title': 'Links to',
                'description': lambda value: list(value),
                'unit': 'links',
                'category': 'SEO',
            },
            'total.number.links': {
                'title': 'Link count',
                'description': lambda value: "This page has %d outbound links." % (value or 0),
                'category': 'HTTP',
            },
            'total.number.invalid_links': {
                'title': 'Total invalid links',
                'description': lambda value: "This page has %d invalid links." % (value or 0),
                'category': 'Semantics',
            },
            'page.invalid_links': {
                'title': 'Invalid Links',
                'description': lambda value: list(value),
                'category': 'Semantics',
            }
        }

    def looks_like_image(self, url):
        image_types = ['png', 'webp', 'gif', 'jpg', 'jpeg']
        for image_type in image_types:
            if url.endswith(image_type):
                return True

        return False

    def get_facts(self):
        links = self.get_links()

        self.review.data['page.links'] = set()
        self.review.data['page.all_links'] = links

        self.add_fact(
            key='page.links',
            value=set(),
        )

        invalid_links = set()

        num_links = 0

        links_to_get = set()

        for link in links:
            url = link.get('href').strip()
            url = REMOVE_HASH.sub('', url)

            if not url:
                continue

            aux = self.normalize_url(url)
            if not aux:
                invalid_links.add(url)
                continue

            url = aux

            if self.looks_like_image(url):
                continue

            if link.get('rel') == 'nofollow':
                continue

            if count_url_levels(url) > self.config.MAX_URL_LEVELS:
                logging.info('Max URL levels! Details: %s' % url)
                continue

            should_get = False
            domain, domain_url = get_domain_from_url(url)
            if domain in self.page_url:
                should_get = True

            if should_get and URL_RE.match(url):
                num_links += 1
                links_to_get.add(url)

        for url in links_to_get:
            self.async_get(url, self.handle_url_loaded)

        self.add_fact(
            key='total.number.links',
            value=num_links
        )

        self.add_fact(
            key='total.number.invalid_links',
            value=len(invalid_links)
        )

        self.add_fact(
            key='page.invalid_links',
            value=invalid_links
        )

    def handle_url_loaded(self, url, response):
        logging.debug('Got response (%s) from %s!' % (response.status_code, url))
        self.review.facts['page.links']['value'].add(url)
        self.review.data['page.links'].add((url, response))

    def get_links(self):
        return self.reviewer.current_html.cssselect(':not(script) a[href]')



#!/usr/bin/python
# -*- coding: utf-8 -*-


from holmes.validators.base import Validator


class GoogleAnalyticsValidator(Validator):

    @classmethod
    def get_analytics_message(cls, value=None):
        return 'This page should include a Google Analytics.'

    @classmethod
    def get_account_message(cls, value=None):
        return 'This page should include a Google Analytics account.'

    @classmethod
    def get_domain_message(cls, value=None):
        return 'This page should include a Google Analytics domain.'

    @classmethod
    def get_violation_definitions(cls):
        return {
            'google_analytics.not_found': {
                'title': 'Google Analytics not found',
                'description': cls.get_analytics_message,
                'category': 'SEO'
            },
            'google_analytics.account.not_found': {
                'title': 'Google Analytics account not found',
                'description': cls.get_account_message,
                'category': 'SEO'
            },
            'google_analytics.domain.not_found': {
                'title': 'Google Analytics domain not found',
                'description': cls.get_domain_message,
                'category': 'SEO'
            }
        }

    def validate(self):
        google_analytics = self.get_analytics()

        if not google_analytics:
            self.add_violation(
                key='google_analytics.not_found',
                value=None,
                points=100
            )
            return

        for account, domain in google_analytics:
            if not account:
                self.add_violation(
                    key='google_analytics.account.not_found',
                    value=None,
                    points=50
                )

            if not domain:
                self.add_violation(
                    key='google_analytics.domain.not_found',
                    value=None,
                    points=50
                )

    def get_analytics(self):
        return self.review.data.get('page.google_analytics', None)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
import logging

from uuid import uuid4
from datetime import datetime

from ujson import dumps
import sqlalchemy as sa
from sqlalchemy.orm import relationship

from holmes.models import Base


class Review(Base):
    __tablename__ = "reviews"

    id = sa.Column(sa.Integer, primary_key=True)
    is_active = sa.Column('is_active', sa.Boolean, default=False, nullable=False)
    is_complete = sa.Column('is_complete', sa.Boolean, default=False, nullable=False)
    uuid = sa.Column('uuid', sa.String(36), default=uuid4, nullable=False)

    created_date = sa.Column('created_date', sa.DateTime, default=datetime.utcnow, nullable=False)
    completed_date = sa.Column('completed_date', sa.DateTime, nullable=True)
    completed_day = sa.Column('completed_day', sa.Date, nullable=True)

    failure_message = sa.Column('failure_message', sa.String(2000), nullable=True)

    domain_id = sa.Column('domain_id', sa.Integer, sa.ForeignKey('domains.id'))
    page_id = sa.Column('page_id', sa.Integer, sa.ForeignKey('pages.id'))

    facts = relationship("Fact", cascade="all,delete")
    violations = relationship("Violation", cascade="all,delete")

    def to_dict(self, fact_definitions, violation_definitions):
        return {
            'page': self.page and self.page.to_dict() or None,
            'domain': self.domain and self.domain.name or None,
            'isComplete': self.is_complete,
            'uuid': str(self.uuid),
            'createdAt': self.created_date,
            'completedAt': self.completed_date,
            'facts': [fact.to_dict(fact_definitions) for fact in self.facts],
            'violations': [violation.to_dict(violation_definitions) for violation in self.violations]
        }

    def __str__(self):
        return str(self.uuid)

    def __repr__(self):
        return str(self)

    def add_fact(self, key, value):
        if self.is_complete:
            raise ValueError("Can't add anything to a completed review.")

        from holmes.models.fact import Fact  # to avoid circular dependency

        fact = Fact(key=key, value=value)

        self.facts.append(fact)

    def add_violation(self, key, value, points, domain):
        if self.is_complete:
            raise ValueError("Can't add anything to a completed review.")

        from holmes.models.violation import Violation  # to avoid circular dependency

        violation = Violation(
            key=key,
            value=value,
            points=int(float(points)),
            domain=domain
        )

        self.violations.append(violation)

    @property
    def failed(self):
        return self.failure_message is not None

    @classmethod
    def get_last_reviews(cls, db, limit=12):
        return db.query(Review).filter(Review.is_active == True) \
                               .order_by(Review.completed_date.desc())[:limit]

    @classmethod
    def get_reviews_count_in_period(cls, db, from_date, to_date=None):
        if to_date is None:
            to_date = datetime.utcnow()

        reviews = db \
            .query(Review.completed_date) \
            .filter(Review.is_active == True) \
            .filter(Review.completed_date.between(from_date, to_date)) \
            .order_by(Review.completed_date.asc()) \
            .all()

        count = len(reviews)

        first_date = None
        if count > 0:
            first_date = reviews[0].completed_date

        return count, first_date

    def get_violation_points(self):
        points = 0
        for violation in self.violations:
            points += violation.points
        return points

    @classmethod
    def by_uuid(cls, uuid, db):
        return db.query(Review).filter(Review.uuid == uuid).first()

    @property
    def violation_count(self):
        return len(self.violations)

    @classmethod
    def _filter_violation_key_name(cls, db, query, key_id, domain_filter=None, page_filter=None):
        from holmes.models.domain import Domain  # to avoid circular dependency
        from holmes.models.page import Page  # to avoid circular dependency
        from holmes.models.violation import Violation  # to avoid circular dependency

        query = query \
            .filter(Page.id == Review.page_id) \
            .filter(Violation.review_id == Review.id) \
            .filter(Review.is_active == 1) \
            .filter(Violation.key_id == key_id)

        page_filter_prefix = '%'
        if domain_filter:
            domain = Domain.get_domain_by_name(domain_filter, db)
            if domain:
                query = query.filter(Review.domain_id == domain.id)
                page_filter_prefix = domain.url

        if page_filter:
            query = query.filter(
                Page.url.like(
                    '{0}{1}%'.format(page_filter_prefix, page_filter)
                )
            )

        return query

    @classmethod
    def count_by_violation_key_name(cls, db, key_id, domain_filter=None, page_filter=None):
        from holmes.models.review import Review  # to avoid circular dependency
        from holmes.models.violation import Violation  # to avoid circular dependency

        query = db \
            .query(sa.func.count(Violation.id)) \
            .filter(Violation.review_is_active == 1) \
            .filter(Violation.key_id == key_id)

        page_filter_prefix = '%'
        if domain_filter:
            from holmes.models.domain import Domain  # to avoid circular dependency
            domain = Domain.get_domain_by_name(domain_filter, db)
            if domain:
                query = query.filter(Review.domain_id == domain.id)
                page_filter_prefix = domain.url

        if page_filter:
            from holmes.models.page import Page  # to avoid circular dependency
            query = query.filter(Review.id == Violation.review_id) \
                .filter(Page.id == Review.page_id) \
                .filter(
                    Page.url.like(
                        '{0}{1}%'.format(page_filter_prefix, page_filter)
                    )
                )

        return query.scalar()

    @classmethod
    def get_by_violation_key_name(cls, db, key_id, current_page=1, page_size=10, domain_filter=None, page_filter=None):
        from holmes.models.page import Page  # to avoid circular dependency
        from holmes.models.review import Review  # to avoid circular dependency
        from holmes.models.violation import Violation  # to avoid circular dependency

        lower_bound = (current_page - 1) * page_size
        upper_bound = lower_bound + page_size

        query = db \
            .query(
                Review.uuid.label('review_uuid'),
                Page.url,
                Page.uuid.label('page_uuid'),
                Review.completed_date
            ) \
            .filter(Violation.key_id == key_id) \
            .filter(Review.id == Violation.review_id) \
            .filter(Review.is_active == 1) \
            .filter(Page.id == Review.page_id) \

        page_filter_prefix = '%'
        if domain_filter:
            from holmes.models.domain import Domain  # to avoid circular dependency
            domain = Domain.get_domain_by_name(domain_filter, db)
            if domain:
                query = query.filter(Review.domain_id == domain.id)
                page_filter_prefix = domain.url

        if page_filter:
            query = query.filter(
                Page.url.like(
                    '{0}{1}%'.format(page_filter_prefix, page_filter)
                )
            )

        return query.order_by(Review.completed_date.desc())[lower_bound:upper_bound]

    @classmethod
    def save_review(cls, page_uuid, review_data, db, fact_definitions, violation_definitions, cache, publish, config):
        from holmes.models import Page

        page = Page.by_uuid(page_uuid, db)
        last_review = page.last_review

        review = Review(
            domain_id=page.domain.id,
            page_id=page.id,
            is_active=True,
            is_complete=False,
            completed_date=datetime.utcnow(),
            uuid=uuid4(),
        )

        db.add(review)

        for fact in review_data['facts']:
            name = fact['key']
            key = fact_definitions[name]['key']
            review.add_fact(key, fact['value'])

        for violation in review_data['violations']:
            name = violation['key']
            key = violation_definitions[name]['key']
            review.add_violation(
                key,
                violation['value'],
                violation['points'],
                page.domain
            )

        page.expires = review_data['expires']
        page.last_modified = review_data['lastModified']
        page.last_review_uuid = review.uuid
        page.last_review = review
        page.last_review_date = review.completed_date
        page.violations_count = len(review_data['violations'])

        review.is_complete = True

        if not last_review:
            cache.increment_active_review_count(page.domain)

            cache.increment_violations_count(
                page.domain,
                increment=page.violations_count
            )

            cache.increment_next_jobs_count(-1)

        else:
            old_violations_count = len(last_review.violations)
            new_violations_count = len(review.violations)

            cache.increment_violations_count(
                page.domain,
                increment=new_violations_count - old_violations_count
            )

            for violation in last_review.violations:
                violation.review_is_active = False

            last_review.is_active = False

        Review.delete_old_reviews(db, config, page)

        publish(dumps({
            'type': 'new-review',
            'reviewId': str(review.uuid)
        }))

    @classmethod
    def delete_old_reviews(cls, db, config, page):
        reviews = db \
            .query(Review) \
            .filter(Review.is_active == 0) \
            .filter(Review.page_id == page.id) \
            .order_by(Review.completed_date.desc()) \
            .all()

        last_reviews = reviews[config.NUMBER_OF_REVIEWS_TO_KEEP:]

        if len(last_reviews) > 0:
            for review in last_reviews:
                db.delete(review)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from tornado.gen import coroutine

from holmes.models import Domain, Request
from holmes.handlers import BaseHandler


class DomainsHandler(BaseHandler):

    @coroutine
    def get(self):
        domains = self.db.query(Domain).order_by(Domain.name.asc()).all()

        if not domains:
            self.write("[]")
            return

        result = []

        for domain in domains:
            result.append({
                "id": domain.id,
                "url": domain.url,
                "name": domain.name,
                "is_active": domain.is_active
            })

        self.write_json(result)


class DomainsFullDataHandler(BaseHandler):

    @coroutine
    def get(self):
        result = self.girl.get('domains_details')
        self.write_json(result)


class DomainDetailsHandler(BaseHandler):

    @coroutine
    def get(self, domain_name):
        domain = Domain.get_domain_by_name(domain_name, self.db)

        if not domain:
            self.set_status(404, 'Domain %s not found' % domain_name)
            return

        result = self.girl.get('domains_details')
        data = next((l for l in result if l['name'] == domain_name), None)

        if not data:
            self.set_status(404, 'Domain %s not found' % domain_name)
            return

        page_count = data.get('pageCount', 0)
        review_count = data.get('reviewCount', 0)
        violation_count = data.get('violationCount', 0)
        error_percentage = data.get('errorPercentage', 0)
        response_time_avg = data.get('averageResponseTime', 0)
        review_percentage = data.get('reviewPercentage', 0)

        domain_json = {
            "id": domain.id,
            "name": domain.name,
            "url": domain.url,
            "pageCount": page_count,
            "reviewCount": review_count,
            "violationCount": violation_count,
            "reviewPercentage": review_percentage,
            "is_active": domain.is_active,
            "errorPercentage": error_percentage,
            "averageResponseTime": response_time_avg,
        }

        self.write_json(domain_json)


class DomainViolationsPerDayHandler(BaseHandler):

    def get(self, domain_name):
        domain = Domain.get_domain_by_name(domain_name, self.db)

        if not domain:
            self.set_status(404, 'Domain %s not found' % domain_name)
            return

        violations_per_day = domain.get_violations_per_day(self.db)

        domain_json = {
            "id": domain.id,
            "name": domain.name,
            "url": domain.url,
            "violations": violations_per_day
        }

        self.write_json(domain_json)


class DomainReviewsHandler(BaseHandler):

    @coroutine
    def get(self, domain_name):
        term = self.get_argument('term', None)
        current_page = int(self.get_argument('current_page', 1))
        page_size = int(self.get_argument('page_size', 10))

        domain = Domain.get_domain_by_name(domain_name, self.db)

        if not domain:
            self.set_status(404, 'Domain %s not found' % domain_name)
            return

        reviews = domain.get_active_reviews(
            self.db,
            url_starts_with=term,
            current_page=current_page,
            page_size=page_size
        )

        if term:
            review_count = len(reviews)
        else:
            review_count = yield self.cache.get_active_review_count(domain)

        result = {
            'reviewCount': review_count,
            'pages': [],
        }

        for page in reviews:
            result['pages'].append({
                "url": page.url,
                "uuid": str(page.uuid),
                "violationCount": page.violations_count,
                "completedAt": page.last_review_date,
                "reviewId": str(page.last_review_uuid)
            })

        self.write_json(result)


class DomainGroupedViolationsHandler(BaseHandler):

    @coroutine
    def get(self, domain_name):
        domain = Domain.get_domain_by_name(domain_name, self.db)

        if not domain:
            self.set_status(404, 'Domain %s not found' % domain_name)
            return

        violation_defs = self.application.violation_definitions

        grouped_violations = self.girl.get('violation_count_by_category_for_domains')

        total = 0
        violations = []

        for item in grouped_violations.get(domain.id, []):
            key_name, key_category_id, count = item['key_name'], item['category_id'], item['violation_count']
            violations.append({
                'categoryId': key_category_id,
                'categoryName': violation_defs[key_name]['category'],
                'count': count
            })
            total += count

        result = {
            "domainId": domain.id,
            'domainName': domain.name,
            'domainURL': domain.url,
            'total': total,
            'violations': violations
        }

        self.write_json(result)


class DomainTopCategoryViolationsHandler(BaseHandler):

    @coroutine
    def get(self, domain_name, key_category_id):
        domain = Domain.get_domain_by_name(domain_name, self.db)

        if not domain:
            self.set_status(404, 'Domain %s not found' % domain_name)
            return

        violation_defs = self.application.violation_definitions

        top_violations = yield self.cache.get_top_in_category_for_domain(
            domain,
            key_category_id,
            self.application.config.get('TOP_CATEGORY_VIOLATIONS_LIMIT')
        )

        violations = []
        for key_name, count in top_violations:
            violations.append({
                'title': violation_defs[key_name]['title'],
                'count': count
            })

        result = {
            "domainId": domain.id,
            'domainName': domain.name,
            'domainURL': domain.url,
            'categoryId': key_category_id,
            'violations': violations
        }

        self.write_json(result)


class DomainsChangeStatusHandler(BaseHandler):

    @coroutine
    def post(self, domain_name):
        domain = Domain.get_domain_by_name(domain_name, self.db)

        if not domain:
            self.set_status(404, 'Domain %s not found' % domain_name)
            return

        domain.is_active = not domain.is_active

        if not domain.is_active:
            yield self.cache.delete_limit_usage_by_domain(domain.url)

    @coroutine
    def options(self, domain_name):
        super(DomainsChangeStatusHandler, self).options()



#!/usr/bin/python
# -*- coding: utf-8 -*-

import re

from holmes.validators.base import Validator

REMOVE_HASH = re.compile('([#].*)$')


class LinkCrawlerValidator(Validator):
    def __init__(self, *args, **kw):
        super(LinkCrawlerValidator, self).__init__(*args, **kw)
        self.broken_links = set()
        self.moved_links = set()

    @classmethod
    def get_redirect_message(cls, value):
        return 'A link from your page to "%s" is using a 302 or 307 redirect. ' \
            'It passes 0%% of link juice (ranking power) and, in most cases, should not be used. ' \
            'Use 301 instead.' % (
                ', '.join(
                    [
                        '<a href="%s" target="_blank">Link #%s</a>' % (url, index)
                        for index, url in enumerate(value)
                    ]
                )
            )

    @classmethod
    def get_broken_link_message(cls, value):
        return 'This page contains broken links to %s ' \
            '(the urls failed to load or timed-out after 10 seconds). ' \
            'This can lead your site to lose rating with ' \
            'Search Engines and is misleading to users.' % (
                ', '.join(
                    [
                        '<a href="%s" target="_blank">Link #%s</a>' % (url, index)
                        for index, url in enumerate(value)
                    ]
                )
            )

    @classmethod
    def get_violation_definitions(cls):
        return {
            'link.broken': {
                'title': 'Broken link(s) found',
                'description': cls.get_broken_link_message,
                'category': 'HTTP'
            },
            'link.moved.temporarily': {
                'title': 'Moved Temporarily',
                'description': cls.get_redirect_message,
                'category': 'HTTP'
            }
        }

    def validate(self):
        links = self.get_links()

        total_score = float(self.reviewer.page_score)
        tax = total_score * float(self.reviewer.config.PAGE_SCORE_TAX_RATE)
        available_score = total_score - tax

        number_of_links = float(len(links)) or 1.0
        link_score = available_score / number_of_links

        for url, response in links:
            self.send_url(response.effective_url, link_score, response)

        if self.broken_links:
            self.add_violation(
                key='link.broken',
                value=self.broken_links,
                points=100 * len(self.broken_links)
            )

        if self.moved_links:
            self.add_violation(
                key='link.moved.temporarily',
                value=self.moved_links,
                points=100
            )

        self.flush()

    def flush(self, *args, **kw):
        super(LinkCrawlerValidator, self).flush(*args, **kw)
        self.broken_links = set()
        self.moved_links = set()

    def get_links(self):
        return self.review.data['page.links']

    def broken_link_violation(self, url, response):
        self.broken_links.add(url)

    def moved_link_violation(self, url, response):
        self.moved_links.add(url)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging
import sqlalchemy as sa
from ujson import loads
import datetime
from tornado.concurrent import return_future

from holmes.models import Base


class User(Base):
    __tablename__ = "users"

    id = sa.Column(sa.Integer, primary_key=True)
    fullname = sa.Column('fullname', sa.String(200), nullable=False)
    email = sa.Column('email', sa.String(100), nullable=False, unique=True)
    is_superuser = sa.Column(
        'is_superuser',
        sa.Boolean,
        nullable=False,
        default=False
    )
    last_login = sa.Column('last_login', sa.DateTime, nullable=True)

    def to_dict(self):
        return {
            'fullname': self.fullname,
            'email': self.email,
            'is_superuser': bool(self.is_superuser),
            'last_login': self.last_login
        }

    def __str__(self):
        return str(self.email)

    def __repr__(self):
        return str(self)

    @classmethod
    def by_email(cls, email, db):
        return db.query(User).filter(User.email == email).first()

    @classmethod
    @return_future
    def authenticate(cls, access_token, fetch_method, db, config, callback):
        logging.info('Authenticating...')

        google_api_url = 'https://www.googleapis.com/oauth2/v1/tokeninfo'
        url = '%s?access_token=%s' % (google_api_url, access_token)

        fetch_method(
            url,
            cls.handle_authenticate(cls.handle_authorize(db, config, callback)),
            proxy_host=config.HTTP_PROXY_HOST,
            proxy_port=config.HTTP_PROXY_PORT
        )

    @classmethod
    def handle_authenticate(cls, callback):
        def handle(*args, **kw):
            response = args[-1]
            callback(response.code, response.body)
        return handle

    @classmethod
    def handle_authorize(cls, db, config, callback):
        def handle(code, body):
            if code > 399:
                callback(({
                    'reason': 'Error',
                    'status': code,
                    'details': body
                }))
                return

            data = loads(body)

            # Verify that the access token is valid for this app.
            if data.get('issued_to') != config.GOOGLE_CLIENT_ID:
                callback({
                    'status': 401,
                    'reason': "Token's client ID does not match app's.",
                })
                return

            user_email = data.get('email')

            from holmes.models import User

            user = User.by_email(user_email, db)
            if user:
                user.last_login = datetime.datetime.now()
                db.flush()
                db.commit()
                callback({
                    'status': 200,
                    'user': user.to_dict()
                })
                return

            callback({
                'status': 403,
                'reason': 'Unauthorized user'
            })
            return

        return handle



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.models import Settings
from holmes.handlers import BaseHandler


class TaxHandler(BaseHandler):
    def post(self):
        tax = float(self.get_argument('tax'))

        self.db.query(Settings).update({'lambda_score': Settings.lambda_score + tax})

        self.write('OK')
        self.finish()



"""domain_is_active

Revision ID: 7fef2bae291
Revises: 531d795ea74
Create Date: 2014-01-17 18:44:04.637011

"""

# revision identifiers, used by Alembic.
revision = '7fef2bae291'
down_revision = '531d795ea74'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        'domains',
        sa.Column('is_active', sa.Boolean, default=True, nullable=False)
    )
    connection = op.get_bind()
    connection.execute('UPDATE domains set is_active=1;')


def downgrade():
    op.drop_column('domains', 'is_active')



#!/usr/bin/python
# -*- coding: utf-8 -*-

from holmes.validators.base import Validator


class MetaTagsValidator(Validator):

    @classmethod
    def get_violation_definitions(cls):
        return {
            'absent.metatags': {
                'title': 'Meta tags not present',
                'description': lambda value: (
                    "No meta tags found on this page. This is damaging for "
                    "Search Engines."
                ),
                'category': 'HTTP'
            },
            'page.metatags.description_too_big': {
                'title': 'Maximum size of description meta tag',
                'description': lambda value: (
                    "The meta description tag is longer than %s characters. "
                    "It is best to keep meta descriptions shorter for better "
                    "indexing on Search engines."
                ) % value['max_size'],
                'category': 'SEO'
            }
        }

    def validate(self):
        meta_tags = self.review.data.get('meta.tags', None)

        if not meta_tags:
            self.add_violation(
                key='absent.metatags',
                value='No metatags.',
                points=100
            )

        max_size = self.config.METATAG_DESCRIPTION_MAX_SIZE
        for mt in meta_tags:
            if mt['key'] == 'description' and len(mt['content']) > max_size:
                self.add_violation(
                    key='page.metatags.description_too_big',
                    value={'max_size': max_size},
                    points=20
                )
                break
